// Buttons - Active, Cancel, Close, Action
#ifndef BUTTON_HPP
#define BUTTON_HPP

#ifndef CLUSTER_HPP
#include "cluster.hpp"
#endif
#ifndef COMMAND_HPP
#include "command.hpp"
#endif

class ButtonController;

/*
The basis for all the different kinds of Button.
The name of the Button may contain a tilde (~) character to
indicate that the following character is the quick selection
key for the Button. The tilde is not shown on the screen, and
the quick key is highlighted.
*/
class Button : public Control {
friend class ButtonController;
	ButtonController *controller;

public:
	Button(const char *n,Keys hotkey = Knone,int on = 0);
	// The Button is just large enough to show the string.
	Button(const coord w,const coord h,const char *n,Keys hotkey = Knone,
		int on = 0);
	// The size of the Button is specified in the constructor.

	int pressed() const { return (testFlag(SFselected) != 0); }
	// Returns 1 if the Button has been pressed.
	void press() { setFlag(SFselected); dirty(); }
	// Press the Button - sets the pressed flag, and redraws.
	void unpress() { clearFlag(SFselected); dirty(); }
	// Unpress the Button - resets the pressed flag, and redraws.
	int quickkey() const { return quick; }
	Keys hotkey() const { return hot; }

	virtual int setpressed();
	// User has clicked on button - take any necessary action and return result.

	virtual int leftdown(MouseEvent& e);
	virtual int leftup(MouseEvent& e);
	virtual int mouseenter(MouseEvent& e);
	virtual int mouseexit(MouseEvent& e);
	virtual int dohotkey(KeyboardEvent& e);
	virtual char *inspectinfo(char *buffer = 0,const char *s = 0) const;
	virtual char *debuginfo(char *buffer = 0) const;
private:
	Keys hot;
	int quick;
	};
declare(OCollection,Button);
declare(OIterator,Button);

/*
Class which controls a list of Buttons. 
Used when the Buttons must interact in some way (e.g. RadioButtons).
*/
class ButtonController {
	virtual void setController(Button *b);
	void setController(const OCollection(Button)& c);
protected:
	OCollection(Button) btns;	// The Buttons being controlled.
public:
	ButtonController(int hintSize = 0);
	ButtonController(const OCollection(Button)& c);
	ButtonController& operator= (const OCollection(Button)& c) 
		{ btns = c; setController(btns); return *this; } 
	ButtonController& operator+ (const OCollection(Button) &c)
		{ add(c); return *this; }
	ButtonController& operator+ (const OCollection(Button) *c)
		{ btns + c; setController(*c); return *this; }
	ButtonController& operator+ (Button *c)
		{ add(c); return *this; }
	void add(const OCollection(Button) &c)
		{ btns + c; setController(c); }
	int add(Button *item)
		{ return (btns.add(item) >= 0) ? setController(item), 1 : 0; }
	int size() const { return btns.size(); }
	virtual int buttonPressed(Button *b);
	// Called by a Button in the controller when it is pressed.
	virtual int kbdevent(KeyboardEvent& e);
	};

/*
A specific ButtonController which ensures that only one button is
pressed at a time. Used for RadioButtons.
*/
class OneButtonController : public ButtonController {
public:
	OneButtonController(int hintSize = 0);
	OneButtonController(const OCollection(Button) &c, 
		int whichone = 0);
	
	Button *selected() const { return theone; };
	// Return the currently pressed Button.
	int selectedIndex() const { return btns.elementNo(theone); }
	// Return the item number of the currently pressed Button.
	int setSelection(Button *b)
		{ return (btns.elementNo(b) < 0) ? 0 : (b == theone) ? 1 : b->setpressed(); }
	// Make the specified Button the one pressed.
	// Returns 0 if this couldn't be done for any reason.
	int setSelectionIndex(int i)
		{ Button *b = btns[i]; return b ? ((b == theone) ? 1 : b->setpressed()) : 0; }
	// Make Button number i the one pressed.

	virtual int buttonPressed(Button *b);
protected:
	Button *theone;
	};

/*
A Button which performs some function when clicked. The action is
passed as a pointer to a Command. The Command will be executed
whenever it is clicked, with the particular instance of Button
passed as its doit parameter. This class is subclassed to provide
the ActiveButton - such as is found in dialog boxes. 
*/
class ActionButton : public Button {
public:
	ActionButton(Command *cmd = 0,
		const char *n = "ActionButton",Keys hotkey = Knone);
	// The	size  of  the  ActionButton is just  big  enough  to
	// contain the name.
	ActionButton(coord width,coord height,Command *cmd = 0,
		const char *n = "ActionButton",Keys hotkey = Knone);
	// The	initial size of the ActionButton is specified in the
	// constructor.

	~ActionButton();		// Also deletes the command.

	virtual int setpressed();
	virtual int doit();

	virtual coord minwidth() const;
	virtual void paint(Rectangle r);
	virtual int kbdevent(KeyboardEvent& e);

private:
	Command *command;
	};

// This enum is used when creating ActiveButtons, to indicate whether they
// are to be the default button. 
enum ButtonDefault { notDefault, Default };

/*
An ActionButton which has a border round it. The border is normally
a single box, but shows as a double box when the ActiveButton is the
default (i.e. is selected by default when <cr> is pressed). 
The name of the Button may contain a tilde (~) character to
indicate that the following character is the quick selection
key for the Button. The tilde is not shown on the screen, and
the quick key is highlighted.
*/
class ActiveButton : public ActionButton {
public:
	ActiveButton(Command *cmd = 0,
		const char *n = "ActiveButton",Keys hotkey = Knone,
		ButtonDefault makedefault = notDefault);
	// The size of the ActionButton is just big enough to contain the name.
	ActiveButton(const coord w,const coord h,
		Command *cmd = 0,const char *n = "ActiveButton",
		Keys hotkey = Knone,ButtonDefault makedefault = notDefault);
	// The initial size of the ActionButton is specified in the constructor. 

	void makedefault(ButtonDefault d = Default);
	// This method can be called to turn default status on and off . 

	virtual void setwindow(Window *w);
	virtual coord minwidth() const;
	virtual coord minheight() const;
	virtual int setfocus(int tabstop = 1);	// 1 if will accept focus, else 0
	virtual void paint(Rectangle r);

protected:
	ButtonDefault def;
	};

/*
An ActiveButton which will close the current Window if it is clicked.
Closing a Window involves calling the Window closeok method (which
normally places a CloseEvent for itself in the event queue and
returns 1). The CloseEvent will remove the Window from the
Application and delete it (also any VObjects within it). 
*/
class CloseButton : public ActiveButton {
public:
	CloseButton(const char *n = "O.K.",Keys hotkey = Knone,
		Command *cmd = 0,
		ButtonDefault makedefault = Default);
	// This constructor makes it easy to change the displayed name of the
	// button, and its hot key. Notice that the button is made the default
	// button unless you provide a notDefault argument. 
	CloseButton(Command *cmd,ButtonDefault makedefault = Default,
		const char *n = "O.K.");
	// This constructor makes it easy to add aditional processing to the
	// CloseButton. The supplied cmd Command will be executed when the
	// button is pressed. Only if it returns != 0 will the close process
	// proceed. 

	virtual int doit();
	};

/*
An ActiveButton which will cancel the current Window if it is
clicked. Cancelling a Window involves calling the Window cancelok
method (which normally places a CloseEvent for itself in the event
queue and returns 1). The CloseEvent will remove the Window from the
Application and delete it (also any VObjects within it).
*/
class CancelButton : public ActiveButton {
public:
	CancelButton(const char *n = "Esc=Cancel",Keys hotkey = Kesc,
		Command *cmd = 0,
		ButtonDefault makedefault = notDefault);
	// This constructor makes it easy to change the displayed name of the
	// button, and its hot key. Notice that there is usually a hot key for
	// this button (Esc). Notice also that the button is not the default
	// button unless you provide a Default argument. 
	CancelButton(Command *cmd,
		ButtonDefault makedefault = notDefault,const char *n = "Esc=Cancel");
	// This constructor makes it easy to add aditional processing to the
	// CancelButton. The supplied Command will be executed when the button
	// is pressed. Only if it returns != 0 will the cancel process proceed.
	// Note that the hotkey (Esc) cannot be changed in this constructor. 
	virtual int doit();
	};

// A Controlled Horizontal Cluster of Buttons.
class ButtonHCluster : public HCluster, public ButtonController {
public:
	ButtonHCluster(const char *name = "ButtonHCluster",
		AlignFlag a = AFexpand, BorderFlag f = BFtabstop);
	ButtonHCluster(const OCollection(Button)& buttons,
		const char *name, AlignFlag a = AFexpand, BorderFlag f = BFtabstop);
	ButtonHCluster(const OCollection(Button)& buttons,
		AlignFlag a = AFexpand, BorderFlag f = BFtabstop);
	void add(Button *button);
	void add(VObject *v) { HCluster::add(v); }
	virtual int kbdevent(KeyboardEvent& e);
	};

// A Controlled Vertical Cluster of Buttons.
class ButtonVCluster : public VCluster, public ButtonController {
public:
	ButtonVCluster(const char *name = "ButtonVCluster",
		AlignFlag a = AFexpand, BorderFlag f = BFtabstop);
	ButtonVCluster(const OCollection(Button)& buttons,
		const char *name, AlignFlag a = AFexpand, BorderFlag f = BFtabstop);
	ButtonVCluster(const OCollection(Button)& buttons,
		AlignFlag a = AFexpand, BorderFlag f = BFtabstop);
	void add(Button *button);
	void add(VObject *v) { VCluster::add(v); }
	virtual int kbdevent(KeyboardEvent& e);
	};

// A support function to print a string containing a tilde (~) 
// without printing the tilde, but with the character following 
// the tilde highlighted. Used by Buttons, Menus etc.
extern void paintWithoutTilde(Window *w,const char *s,int color,Point p,
		const Rectangle &clip);

// A support function to determine the printable length of a 
// string containing a tilde (~) which will be printed by 
// paintWithoutTilde. Used by Buttons, Menus etc.
extern int lenWithoutTilde(const char *s);

// A support function to determine if a string contains a tilde (~).
// If it does, the character following the tilde is returned.
// Used by Buttons, Menus etc. to determine the quick selection key.
extern int checkForTilde(const char *s);

#endif
