/*
 * DocClass - class documentation utility
 * Compiler               Command line
 * Zortech C++            ztc -mli docclass.cpp lib\_mainl.obj
 * Borland C++            bcc -ml -N docclass.cpp lib\wildargs.obj
 * GNU C++                g++ docclass.C -o docclass
 * Other compilers
 * Define the following ...
 * STRICMP_PROVIDED         If the library contains the stricmp function
 * STRLWR_PROVIDED          If the platform has case insensitive file names
 *                          and the library has a strlwr function
 * CASE_SENSITIVE_FILENAMES If the platform has case-sensitive file names
 *                          (e.g. Unix)
 * 
 * If you port docclass to another compiler, please add a new compiler-
 * specific section (see below), and mail the section to me at
 * docclass@trmphrst.demon.co.uk. Likewise bug fixes and problems.
 *
 * This code is Copyright (c) 1993 Trumphurst Ltd.
 * Trumphurst Ltd. reserve the right to withdraw the code from public 
 * distribution at any time. A licence is required for commercial use -
 * mail docclass@trmphrst.demon.co.uk for details.
 */

#ifdef __ZTC__			/* Zortech C++ uses .hpp for headers */
#include <iomanip.hpp>
#include <fstream.hpp>
#include <strstream.hpp>
#else
#include <fstream.h>
#include <iomanip.h>
#include <strstream.h>
#endif
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include "docclass.hpp"

/*
 * Compiler specific sections. One for each supported compiler.
 * New compiler specific sections may be inserted here.
 */
// Zortech C++
#ifdef __ZTC__
#define STRLWR_PROVIDED
#endif

// Borland C++
#ifdef __TURBOC__
#define STRICMP_PROVIDED
#define STRLWR_PROVIDED
#endif

// GNU gcc
#ifdef __GNUC__
#define CASE_SENSITIVE_FILENAMES
#endif

// End of compiler specific section

/*
 * $Log:	docclass.cpp $
Revision 1.8  93/11/22  18:20:33  nikki
Fixed Borland & Sun problems. Added OutputProcessor.

Revision 1.7  93/11/15  18:11:30  nikki
Fix some GNU contravariance violations
Improve member parsing for variables whose type was parameterised

 */
static const char RCSrevision[] = "$Revision: 1.8 $";

// Make the Collection array bigger by n elements
int Collection::expand(int n)
{
	const int MAXGRAIN = 5;

	if((uelements + n) > nelements)			// Not enough already ?
		{
		if(n == 1)
			n = MAXGRAIN;					// Allocate a few spare ones
		
		int extras = uelements + n - nelements;	// Extra quantity needed

		int nbytes = (nelements + extras) * sizeof(void *);

		void **l;

		if(list == 0)						// Malloc/realloc the list
			l = (void **)malloc(nbytes);
		else
			l = (void **)realloc(list,nbytes);

		if(!l)
			return 0;
		list = l;
		nelements += extras;				// Store new size
		}
	return 1;
}

Collection::Collection(int hintSize)
: list(0), nelements(0), uelements(0)
{
	if(hintSize > 0)
		expand(hintSize);
}

Collection::Collection(const Collection& c)
: list(0), nelements(0), uelements(0)
{
    *this = c;              // use operator=
}

Collection::~Collection()
{
	if(list)
		free(list);
}

Collection& Collection::operator= (const Collection& c)
{
	if(this == &c)
		return *this;
	nelements = c.nelements;
	uelements = c.uelements;
	if(list)
		free(list);
	if(c.list == 0)
		list = 0;
	else
		{
		int nbytes = nelements * sizeof(void *);
		list = (void **)malloc(nbytes);
		if(list)
			memcpy(list,c.list,nbytes);
		}
	return *this;
}

void Collection::append(void *c)
{
	if(!expand())
		return;
	list[uelements++] = c;
}

Collection& Collection::operator+ (const Collection& c)
{
	expand(c.uelements);
	if(list)
		{
		memcpy(list + uelements,c.list,c.uelements * sizeof(void *));
		uelements += c.uelements;
		}
	return *this;
}

static Collection *sortedCollection;

extern "C" {
static int scompare(const void *v1,const void *v2)
{
	return sortedCollection->compare(*((void **)v1),*((void **)v2));
}
}

void Collection::resort()
{
	if(size() > 0)
		{
		sortedCollection = this;
		qsort(list,size(),sizeof(void**),scompare);
		sortedCollection = 0;
		}
}

int Collection::compare(const void *, const void *) const
{
	return 0;
}

Iterator::Iterator(const Collection& l)
: list(l), i(-1)
{
}

void *Iterator::next()
{
	void *v;

	if(list.size())
		{
		if(i == list.size() - 1)
			{
			i = -1;
			v = 0;
			}
		else
			v = list.list[++i];
		}
	else
		v = 0;
	return v;
}

int indentmult = 4;
int interactive = 1;
int printPrivate = 0;
LineFile inputFile;
int debugMode;

// Get next character, allowing for ungot characters
int LineFile::get()
{
	if(!inputFile)
		return EOF;
	int c = (ungot > lookahead) ? *--ungot : f.get();

	if(c == '\n')
		lno++;
	return c;
}

// Push a character back on the input stream
void LineFile::unget(int c)
{
	if(!inputFile)
		return;
	if(c == '\n')
		lno--;
	if(c != EOF)
		*ungot++ = c;
}

// Skip to next matching string delimiter - " or '
// Allow for escaped delimiters inside the string
// Copy skipped strings to cpos if cpos != 0
void LineFile::skipstring(char delim)
{
	int c;

	while((c = get()) != EOF && c != '\n')
		{
		if(cpos)
			*cpos++ = c;
		if(c == '\\')
			{
			c = get();
			if(cpos && c != EOF)
				*cpos++ = c;
			}
		else if(c == delim)
			break;
		}
	if(cpos)
		*cpos = 0;
}

// Skip to matching bracket - ( or [ or { or <
// Cope correctly with strings containing brackets, and comments
// Copy skipped characters to cpos if cpos != 0
void LineFile::skipbracket(char bracket)
{
	int nest = 1;
	int cbracket;
	int c;

	switch(bracket)
		{
		case '(':
			cbracket = ')';
			break;
		case '[':
			cbracket = ']';
			break;
		case '{':
			cbracket = '}';
			break;
		case '<':
			cbracket = '>';
			break;
		}

	while((c = get()) != EOF)
		{
		if(cpos)
			*cpos++ = c;
		if(c == bracket)
			nest++;
		else if(c == cbracket)
			{
			if(!--nest)
				break;
			}
		else if(c == '"' || c == '\'')
			skipstring(c);
		else if(c == '/')
			{
			int nextc = get();
			
			if(nextc == '/')
				skipcomment(cpos);
			else if(nextc == '*')
				{
				while(skipmultiComment(cpos))
					;
				}
			else
				{
				unget(nextc);
				continue;
				}
			if(cpos)
				--cpos;
			}
		}
	if(cpos)
		*cpos = 0;
}

// Skip a single line comment, copying comment if copy != 0
void LineFile::skipcomment(char *copy)
{
	int c;

	while((c = get()) != EOF && c != '\n')
		{
		if(copy)
			*copy++ = c;
		}
	if(copy)
		*copy = 0;
}

// Skip one line of a multi-line comment, copying comment if copy != 0
int LineFile::skipmultiComment(char *copy)
{
	int c;
	char *start = copy;

	if(!inMultiComment)
		{
		c = skipspace();
		unget(c);
		}
	while((c = get()) != EOF && c != '\n')
		{
		if(c == '*')
			{
			int nextc = get();
			
			if(nextc == '/')
				{
				c = EOF;
				break;
				}
			unget(nextc);
			}
		if(copy)
			*copy++ = c;
		}
	if(copy)		// remove trailing spaces
		{
		while(copy > start && isspace(copy[-1]))
			copy--;
		*copy = 0;
		}
	return (c != EOF);
}

// Skip white space, copying it to lastspace
int LineFile::skipspace()
{
	int c;
	
	for(char *d = lastspace; (c = get()) != EOF && isspace(c); *d++ = c)
		;
	*d = 0;
	return c;
}

LineFile::LineFile(const char *filename)
: fname(""), cpos(tok), ungot(lookahead), lno(0), inMultiComment(0),
	tokt(TTeof), lasttokt(TTeof), holdtokt(TTeof)
{
	tok[0] = lasttok[0] = holdtok[0] = 0;
	if(*filename)
		open(filename);
}

LineFile::~LineFile()
{
	close();
}

int LineFile::open(const char *fn)
{
	close();
	f.open(fn);
	ungot = lookahead;
	lno = 1;
	tok[0] = lasttok[0] = holdtok[0] = 0;
	inMultiComment = 0;
	tokt = lasttokt = holdtokt = TTeof;
	if(f)
		fname = fn;
	else
		fname = "";
	return !!f;
}

void LineFile::close()
{
	if(*fname)
		{
		f.close();
		lno = 0;
		fname = "";
		}
}

static const char *const keywords[] = {
	"name",
	"number",
	"string",
	"arrayref",
	"char",
	"comment",
	"multiComment",
	"ignore",
	"arglist",
	"class",
	"struct",
	"enum",
	"private",
	"public",
	"protected",
	"static",
	"extern",
	"typedef",
	"const",
	"volatile",
	"virtual",
	"inline",
	"operator",
	0 };

// Get next token. Return token type.
int LineFile::getnext()
{
	int c = skipspace(), nextc;

	*cpos++ = c;
	switch(c)
		{
		case EOF:
			tok[0] = 0;
			return (tokt = TTeof);
		case '"':	// String or character constant
		case '\'':
			skipstring(c);
			return (tokt = (c == '"') ? TTstring : TTchar);
		case '/':	// Might be a comment
			switch(nextc = get())
				{
				case '/':	// Single line comment
					*cpos++ = nextc;
					skipcomment(cpos);
					if(!strcmp(tok, "//!"))
						return tokt = TTignore;
					return tokt = TTcomment;
				case '*':	// Multi line comment
					cpos--;			// remove '/' and ignore '*'
					inMultiComment = skipmultiComment(cpos);
					if(!inMultiComment && !strcmp(tok, "!"))
						return tokt = TTignore;
					return tokt = TTmultiComment;
				}
			unget(nextc);	// Not a comment - return '/'
			*cpos = 0;
			return tokt = c;
		case '(':	// Argument list
			skipbracket();
			return tokt = TTarglist;
		case '[':	// Array element specifier
			skipbracket('[');
			return tokt = TTarrayref;
		case '0':	// Numeric
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			while((c = get()) != EOF && isdigit(c))
				*cpos++ = c;
			unget(c);
			*cpos = 0;
			return tokt = TTnumber;
		}
	if(!isalpha(c) && c != '_')
		{
		*cpos = 0;		// Must be punctuation
		return tokt = c;
		}
	// Variable name or keyword
	while((c = get()) != EOF && (isalnum(c) || c == '_'))
		*cpos++ = c;
	unget(c);
	*cpos = 0;

	tokt = TTname;
	for(const char * const *k = keywords + TTkeywords - TTbegins; *k; k++)
		if(!strcmp(tok, *k))	// It is in the keyword list
			{
			tokt = (TTbegins + (k - keywords));
			break;
			}
	return tokt;
}

// Debugging output
void printToken()
{
	if(debugMode > 1)
		cout << inputFile.token() << ".";
	if(inputFile.token() < TTeof)
		cout << keywords[inputFile.token() - TTbegins] << ":";
	else
		cout << "TTpunct:";
	cout << (const char *)inputFile << endl;
}

// Get next token, allowing for ungot tokens & multi-line comments
int LineFile::nextToken()
{
	if(tok[0])
		strcpy(lasttok, tok);
	lasttokt = tokt;
	cpos = tok;
	if(!inputFile)
		tokt = TTeof;
	else if(holdtokt != TTeof)
		{
		tokt = holdtokt;
		strcpy(tok, holdtok);
		holdtokt = TTeof;
		}
	else if(inMultiComment)
		{
		inMultiComment = skipmultiComment(cpos);
		if(inMultiComment || tok[0])
			tokt = TTmultiComment;
		else
			tokt = getnext();
		}
	else
		tokt = getnext();
	if(debugMode)
		printToken();
	return tokt;
}

// Having got a name, check for template/generic arguments
int LineFile::getRestOfName()
{
	if(tokt == TTname)
		{
		int c = get();

		if(c == '(' || c == '<')
			{
			cpos = tok + strlen(tok);
			*cpos++ = c;
			skipbracket(c);
			if(debugMode)
				cout << "... " << tok << endl;
			}
		else
			unget(c);
		}
	return tokt;
}

// Get the next name, including any template/generic arguments
int LineFile::nextName()
{
	if(nextToken() == TTname)
		getRestOfName();
	return tokt;
}

// Put a token back on the input stream
void LineFile::ungetToken()
{
	holdtokt = tokt;
	tokt = lasttokt;
	strcpy(holdtok, tok);
	strcpy(tok, lasttok);
}

// Bracket may be ( or [ or < or {
void LineFile::skipToMatchingBracket(char bracket)
{
	cpos = 0;
	skipbracket(bracket);
}

// Strip trailing space, and return remaining string length
static int stripTrailingSpace(char *p)
{
	char *d = p + strlen(p);

	while(d > p && isspace(d[-1]))
		d--;
	*d = 0;
	return (d - p);
}

NameList::NameList()
{
}

NameList::~NameList()
{
}

void NameList::append(const char *m)
{
	if(!contains(m))
		list + (char *)m;
}

void NameList::append(NameList &n)
{
	list + n.list;
}

int NameList::contains(const char *n)
{
	GIterator(char) next(list);
	char *c;

	while((c = next()) != 0)
		if(!strcmp(n, c))
			return 1;
	return 0;
}

void NameList::zero()
{
	while(list.getlast())
		;
}

OutputProcessor::OutputProcessor()
: headingLevel(HLnone), indentLevel(0)
{
	lineBuffer[0] = 0;
}

OutputProcessor::~OutputProcessor()
{
	if(lineBuffer[0])
		endLine();
}

void OutputProcessor::underline(char linechar)
{
	headingLevel = HLnone + linechar;
}

void OutputProcessor::heading(int level)
{
	headingLevel = (level < HLfirst || level > HLlast) ? HLlast : level;
}

void OutputProcessor::includeFile(const char *filename)
{
	cout << "#include " << '"' << filename << '"' << endl;
}

void OutputProcessor::indent(int n)
{
	indentLevel = n;
}

OutputProcessor &OutputProcessor::operator<<(const char *string)
{
	strcat(lineBuffer, string);
	return *this;
}

OutputProcessor &OutputProcessor::operator<<(const LineEnd)
{
	endLine();
	return *this;
}

void OutputProcessor::endLine()
{
	static const char underlines[HLlast+1] = {
	'=',	// HLclass
	'=',	// HLstruct
	'=',	// HLenum
	'=',	// HLstatic
	'=',	// HLtypedef
	'=',	// HLextern
	'-',	// HLbaseClasses
	'-',	// HLderivedClasses
	'-',	// HLpublics
	'-',	// HLprotecteds
	'-',	// HLprivates
	'-',	// HLinheritedPublics
	'-',	// HLinheritedProtecteds
	0,		// HLmember,
	0,		// HLinheritClass,
	'-',	// HLenumValues
	'?'		// HLlast
	};

	if(indentLevel)
		cout << setw(indentLevel) << "";
	cout << lineBuffer;
	if(headingLevel <= HLlast)
		{
		if(underlines[headingLevel])
			headingLevel = HLnone + underlines[headingLevel];
		else
			cout << ':';
		}
	if(headingLevel > HLnone)
		{
		char c = headingLevel - HLnone;

		cout << endl;
		if(indentLevel)
			cout << setw(indentLevel) << "";
		for(int l = strlen(lineBuffer); l--; )
			cout << c;
		}
	cout << endl;
	headingLevel = HLnone;
	indentLevel = 0;
	lineBuffer[0] = 0;
}

OutputProcessor *outputProcessor;

CommentList::CommentList()
{
}

CommentList::~CommentList()
{
	zero();
}

void CommentList::append(const char *m)
{
	char *p = strchr(m, '\n');
	
	if(p)
		*p = 0;
	if(debugMode)
		{
		cout << "*** CommentList[" << this << "]::append" << m << endl;
		if(debugMode > 1)
			print(8);
		}
	list + strdup(m);
}

void CommentList::append(CommentList &n)
{
	if(debugMode)
		{
		cout << "*** CommentList[" << this << "]::append [" << &n << "]..." << endl;
		if(debugMode > 1)
			print(8);
		n.print(4);
		}
	list + n.list;
	while(n.list.getlast() != 0)
		;
}

void CommentList::zero()
{
	char *p;

	if(debugMode)
		cout << "*** CommentList[" << this << "]::zero" << endl;
	while((p = list.getlast()) != 0)
		free(p);
}

const char *skipspace(const char *p)
{
	while(isspace(*p))
		p++;
	return p;
}

void CommentList::print(int indent) const
{
	GIterator(char) next(list);
	const char *p;

	if(debugMode > 1)
		cout << "***CommentList::print[" << (void *)this << "]" << endl;
	while((p = next()) != 0)
		{
		if(debugMode > 1)
			cout << "[" << p << "]";

		const char *c = skipspace(p);

		if(c[0] == '/' && c[1] == '/')
			{
			c += 2;
			if(*c)
				{
				if(ispunct(*c))
					{
					p = skipspace(c + 1);
					outputProcessor->underline(*c);
					*outputProcessor << p << endLine;
					continue;
					}
				else
					p = skipspace(c);
				}
			else
				p = c;
			}
		outputProcessor->indent(indent);
		*outputProcessor << p << endLine;
		}
}

MemberBase::MemberBase(const char *n, CommentList &p, CommentList &c)
: name(strdup(n))
{
	preamble.append(p);
	comments.append(c);
}

MemberBase::~MemberBase()
{
	if(name)
		free(name);
}

Member::Member(int t, const char *n, const char *s, 
	CommentList &p, CommentList &c)
: MemberBase(n, p, c), sig(strdup(s)), type(t)
{
}

Member::~Member()
{
	if(sig)
		free(sig);
}

void Member::print() const
{
	const char *dest = isdestructor() ? "~" : "";

	outputProcessor->heading(HLmember);
	*outputProcessor << dest << namepart() << endLine;
	outputProcessor->indent(4);
	*outputProcessor << storageClass();
	if(ispure())
		*outputProcessor << "pure ";
	if(isinline())
		*outputProcessor << "inline ";
	if(isvirtual())
		*outputProcessor << "virtual ";
	*outputProcessor << signature();
	if(isconst())
		*outputProcessor << " const";
	*outputProcessor << endLine;
	comments.print(4);;
}

EnumMember::EnumMember(const char *n, const char *v, 
	CommentList &p, CommentList &c)
: MemberBase(n, p, c), val(strdup(v))
{
}

EnumMember::~EnumMember()
{
	if(val)
		free(val);
}

void EnumMember::print() const
{
	preamble.print();
	outputProcessor->indent(2);
	*outputProcessor << namepart() << val << endLine;
	comments.print(4);
}

#ifdef __ZTC__
#define stricmp(n1, n2) strcmpl(n1, n2)
#else
#ifndef STRICMP_PROVIDED
int stricmp(const char *s1, const char *s2)
{
	while(*s1 && toupper(*s1) == toupper(*s2))
		{
		s1++;
		s2++;
		}
	return (*s2 - *s1);
}
#endif

#ifndef STRLWR_PROVIDED
#ifndef CASE_SENSITIVE_FILENAMES
void strlwr(char *p)
{
	while(*p)
		{
		*p = tolower(*p);
		p++;
		}
}
#endif	/* !CASE_SENSITIVE_FILENAMES */
#endif  /* !STRLWR_PROVIDED */
#endif  /* __ZTC__ */

int filesort(const ClassBase *b1, const ClassBase *b2)
{
	int retval = strcmp(b1->filename(), b2->filename());

	if(!retval)
		retval = b1->lineno() - b2->lineno();
	return retval;
}

int alphasort(const ClassBase *b1, const ClassBase *b2)
{
	int retval = stricmp(b1->name(), b2->name());

	if(!retval)
		retval = strcmp(b1->name(), b2->name());
	return retval;
}

ClassRef::ClassRef(Class *c, int p)
: info(c), prot(p)
{
}

OutputProcessor &operator<<(OutputProcessor &o, const ClassRef &r)
{
	static char *protectind[] = {
		"public ",
		"protected ",
		"private "
		};

	if(r.isvirtual())
		o << "virtual ";
	o << protectind[r.protection()] << *r.theClass();
	return o;
}

void ClassRef::print(int indent,ClassFunc func) const
{
	outputProcessor->indent(indent + 1);
	*outputProcessor << *this << endLine;
	if(func)
		(info->*func)(indent + indentmult);
}

GCollection(ClassRef) emptyRefs;

ClassBase::ClassBase(const char *n)
: cname(strdup(n)), next(0), 
	fname(inputFile.filename()), lno(inputFile.lineno())
{
	classList.append(this);
}

ClassBase::~ClassBase()
{
	if(cname)
		free(cname);
}

BaseWithMembers::BaseWithMembers(const char *n)
: ClassBase(n)
{
}

BaseWithMembers::~BaseWithMembers()
{
}

void BaseWithMembers::appendMember(MemberBase *m)
{
	if(!members.size())
		{
		fname = inputFile.filename();
		lno = inputFile.lineno();
		}
	members.append(m);
}

Class::Class(const char *n, Type typ)
: BaseWithMembers(n), t(typ)
{
	classOnlyList.append(this);
}

Class::~Class()
{
	ClassRef *b;

	while((b = bases.getlast()) != 0)
		delete b;
}

void Class::appendBase(Class *c, int p)
{
	bases.append(new ClassRef(c,p));
	c->derived.append(new ClassRef(this,p));
}

Class *Class::theClass(const char *name, Type typ)
{
	GIterator(Class) next(classOnlyList);
	Class *c;

	while((c = next()) != 0)
		if(c->type() == typ && !strcmp(c->name(),name))
			return c;
	return new Class(name, typ);
}

void Class::print(int indent,ClassFunc func) const
{
	outputProcessor->indent(indent + 1);
	outputProcessor->heading(HLinheritClass);
	*outputProcessor << *this << endLine;
	if(func)
		(this->*func)(indent + indentmult);
}

void Class::printbases(int indent) const
{
	GIterator(ClassRef) next(bases);
	ClassRef *b;

	while((b = next()) != 0)
		b->print(indent,&Class::printbases);
}

void Class::printderived(int indent) const
{
	GIterator(ClassRef) next(derived);
	ClassRef *b;

	while((b = next()) != 0)
		b->print(indent,&Class::printderived);
}

NameList Class::mynames;
int Class::doingBases = 0;

void Class::printSomeMembers(NameList &names, int type) const
{
	GIterator(MemberBase) nextMember(members);
	Member *m;
	int started = 0;

	// naughty downcast
	while((m = (Member *)nextMember()) != 0)
		{
		if(m->protection() != type)
			continue;
		if(type == MTprivate && !printPrivate && !m->isvirtual())
			continue;			// print virtual private members only
		if(!names.contains(m->namepart()))
			{
			if(doingBases)
				{
				if(m->isstatic() || m->isprivate())
					continue;	// Ignore statics and privates in base classes
				if(!strcmp(m->namepart(), name()))
					continue;	// Ignore constructors and destructors
				if(!strcmp(m->namepart(), "operator="))
					continue;	// operator= is not inherited
				if(mynames.contains(m->namepart()))
					continue;	// Print each inherited member only once
				}
			if(!started)
				{
				static char *heading[] =
					{
					"Public Members",
					"Protected Members",
					"Private Members"
					};

				if(doingBases)
					{
					outputProcessor->heading(HLinheritedPublics + type);
					*outputProcessor << heading[type] << " from " << name() << endLine;
					}
				else
					{
					outputProcessor->heading(HLpublics + type);
					*outputProcessor << heading[type] << endLine;
					}
				started = 1;
				}
			if(doingBases)
				{
				outputProcessor->indent(4);
				*outputProcessor << m->namepart() << endLine;
				}
			else
				m->print();
			mynames.append(m->namepart());
			}
		}

}
void Class::printmembers(NameList &names) const
{
	mynames.zero();
	printSomeMembers(names, MTpublic);
	printSomeMembers(names, MTprotected);
	if(!doingBases)
		printSomeMembers(names, MTprivate);
	names.append(mynames);

	GIterator(ClassRef) next(bases);
	ClassRef *b;

	doingBases = 1;
	while((b = next()) != 0)
		{
		Class *c = (Class *)b->theClass();
		
		c->printmembers(names);
		}
}

void Class::printdoc() const
{
	outputProcessor->heading(type() == Tclass ? HLclass : HLstruct);
	*outputProcessor << typeString() << " " << name() << endLine;
	if(filename()[0])
		outputProcessor->includeFile(filename());
	if(comments.size())
		{
		*outputProcessor << endLine;
		comments.print();
		}
	if(baseClasses().size())
		{
		outputProcessor->heading(HLbaseClasses);
		*outputProcessor << "Base classes" << endLine;
		printbases();
		}
	if(derivedClasses().size())
		{
		outputProcessor->heading(HLderivedClasses);
		*outputProcessor << "Derived classes";
		printderived();
		}
	if(members.size())
		{
		NameList names;
		doingBases = 0;
		printmembers(names);
		}
	*outputProcessor << endLine;
}

Enum::Enum(const char *n)
: BaseWithMembers(n)
{
	enumOnlyList.append(this);
}

Enum::~Enum()
{
}

Enum *Enum::theEnum(const char *name)
{
	GIterator(Enum) next(enumOnlyList);
	Enum *c;

	while((c = next()) != 0)
		if(!strcmp(c->name(),name))
			return c;
	return new Enum(name);
}

void Enum::printmembers() const
{
	GIterator(MemberBase) nextMember(members);
	MemberBase *m;

	// naughty downcast
	while((m = nextMember()) != 0)
		m->print();
}

void Enum::printdoc() const
{
	outputProcessor->heading(HLenum);
	*outputProcessor << "enum " << name() << endLine;
	if(filename()[0])
		outputProcessor->includeFile(filename());
	if(comments.size())
		{
		*outputProcessor << endLine;
		comments.print();
		}
	if(members.size())
		{
		outputProcessor->heading(HLenumValues);
		*outputProcessor << "Values" << endLine;
		printmembers();
		}
	*outputProcessor << endLine;
}

Extern::Extern(int t, const char *n, const char *v, CommentList &p, 
	CommentList &c)
: ClassBase(n), MemberBase(v, p, c), type(t)
{
	c.zero();
	externOnlyList.append(this);
}

Extern::~Extern()
{
}

Extern *Extern::theExtern(int type, const char *name, 
		const char *v, CommentList &p, CommentList &comments)
{
	GIterator(Extern) next(externOnlyList);
	Extern *c;

	while((c = next()) != 0)
		if(!strcmp(c->cname,name))
			{
			p.zero();
			comments.zero();
			return c;
			}
	return new Extern(type, name, v, p, comments);
}

void Extern::printdoc() const
{
	outputProcessor->heading((isstatic() ? HLstatic : 
			(istypedef() ? HLtypedef : HLextern)));
	*outputProcessor << storageClass() << cname << endLine;
	if(filename()[0])
		outputProcessor->includeFile(filename());
	if(preamble.size())
		{
		*outputProcessor << endLine;
		preamble.print();
		}
	*outputProcessor << endLine;
	outputProcessor->indent(4);
	*outputProcessor << storageClass();
	if(isinline())
		*outputProcessor << "inline ";
	*outputProcessor << signature() << endLine;
	comments.print(4);
	*outputProcessor << endLine;
}

void Extern::print() const
{
	printdoc();
}

int collectMember(CommentList &comments, char *mname, char *signature)
{
	int isop = 0;
	int gotargs = 0;
	int discard = 0;
	int gotpure = 0;
	int type = 0;
	int gotignore = 0;
	static char lastName[LINESIZE];

	*mname = 0;
	*signature = 0;
	lastName[0] = 0;
	if(debugMode)
		cout << "*** collectMember" << endl;
	for ( ; ; inputFile.nextToken())
		{
		int lt = inputFile.token();

		switch(lt)
			{
			case TTeof:
			case '}':
				return -1;			// unexpected EOF
			case TTmultiComment:
			case TTcomment:
				comments.append(inputFile);
				continue;
			case TTignore:
				gotignore = 1;
				continue;
			case TTname:
				strcpy(lastName, inputFile);
				if(!isop && !discard)
					*mname = 0;
				gotargs = 0;
				discard = 0;
				break;
			case TTarrayref:
				gotargs = 0;
				discard = 0;
				break;
			case TTstatic:
				type |= MTstatic;
				discard = 0;
				continue;
			case TTtypedef:
				type |= MTtypedef;
				discard = 0;
				continue;
			case TTvirtual:
				type |= MTvirtual;
				discard = 0;
				continue;
			case TTinline:
				type |= MTinline;
				discard = 0;
				continue;
			case TToperator:
				isop = 1;
				gotargs = 0;
				mname[0] = 0;
				discard = 0;
				break;
			case TTarglist:
				if(isop)
					strcpy(lastName, mname);
				else
					strcpy(mname, lastName);
				gotargs = strlen(mname);
				discard = 0;
				break;
			case TTconst:
				discard = strlen(signature);
				gotpure = 0;
				break;
			case '\\':		// ignore
				continue;
			case '{':
				if(gotargs)
					type |= MTinline;
				inputFile.ungetToken();
				/* drop through */
			case ';':
				if(discard)
					{
					if(!strncmp(signature + discard, "const", 5))
						type |= MTconst;
					if(gotpure)
						type |= MTpure;
					signature[discard] = 0;
					}
				stripTrailingSpace(signature);
				if(!*mname)
					strcpy(mname, lastName);
				stripTrailingSpace(mname);
				if(gotargs)
					{
					mname[gotargs] = 0;
					stripTrailingSpace(mname);
					}
				return gotignore ? -1 : type;
			case '~':
				if(!*signature)
					type |= MTdestructor;
				discard = 0;
				break;
			case '=':
				if(!discard)
					discard = strlen(signature);
				break;
			case TTnumber:
				if(inputFile.lasttoken() == '=' && inputFile == "0")
					gotpure = 1;
				break;
			}
		if(*signature)
			strcat(signature, inputFile.space());
		strcat(signature, inputFile);
		if(isop)
			{
			if(*mname)
				strcat(mname, inputFile.space());
			strcat(mname, inputFile);
			}
		}
}

Member *parseMember(int prot, CommentList &l)
{
	if(debugMode)
		cout << "*** parseMember" << endl;

	static char n[LINESIZE], s[LINESIZE];
	CommentList comments;
	int type = collectMember(comments, n, s);

	if(debugMode)
		cout << "*** collectMember returned " << type << endl;
	if(type == -1)
		return 0;
	switch(prot)
		{
		case Protect:
			type |= MTprotected;
			break;
		case Private:
			type |= MTprivate;
			break;
		}
	return new Member(type, n, s, l, comments);
}

Extern *parseExtern(CommentList &preamble, int storageType)
{
	int extC = 0;

	if(inputFile.nextToken() == TTstring)
		{
		extC = 1;
		if(inputFile.nextToken() == '{')
			return 0;		// can't parse extern "C" { ... } yet
		else
			inputFile.ungetToken();
		}

	if(debugMode)
		cout << "*** parseExtern" << endl;

	static char n[LINESIZE], s[LINESIZE];
	CommentList comments;

	int type = collectMember(comments, n, s);

	if(debugMode)
		cout << "*** collectMember returned " << type << ":" << n << endl;
	if(type == -1 || !n[0] || !s[0])
		return 0;
	type &= ~(MTvirtual | MTpure | MTprotected | MTconst | MTdestructor);
	type |= storageType;
	if(extC)
		type |= MTexternC;
	return Extern::theExtern(type, n, s, preamble, comments);
}

void parseBases(Class &current)
{
	int p = Public, v = 0;

	if(debugMode)
		cout << "*** parseBases" << endl;
	for( ; ; )
		{
		switch(inputFile.nextName())
			{
			case ',':
				break;
			case TTpublic:
				break;
			case TTprivate:
				p = Private;
				break;
			case TTprotected:
				p = Protect;
				break;
			case TTvirtual:
				v = Virtual;
				break;
			case TTname:
				current.appendBase(Class::theClass(inputFile), p | v);
				p = Public;
				v = 0;
				break;
			default:
				if(debugMode)
					cout << "*** parseBases returned" << endl;
				return;
			}
		}
}

void finishComment(MemberBase *&member, CommentList &comments, int anyway = 0)
{
	if(member)
		{
		if(debugMode)
			{
			cout << "*** finishComment " <<  member->namepart() << endl;
			comments.print(4);
			}
		member->append(comments);
		comments.zero();
		member = 0;
		}
	else if(anyway && comments.size())
		comments.zero();
}

int collectEnumMember(CommentList &comments, char *mname, char *value)
{
	int invalue = 0;

	*mname = 0;
	*value = 0;
	if(debugMode)
		cout << "*** collectEnumMember" << endl;
	for ( ; ; inputFile.nextToken())
		{
		int lt = inputFile.token();

		switch(lt)
			{
			case TTeof:
				return -1;			// unexpected EOF
			case TTmultiComment:
			case TTcomment:
				comments.append(inputFile);
				continue;
			case TTignore:
				*mname = 0;
				continue;
			case TTname:
				if(!invalue)
					strcpy(mname, inputFile);
				break;
			case TTarrayref:
				break;
			case '}':
				inputFile.ungetToken();
			case ',':
				return (*mname != 0);
			case '=':
				invalue = 1;
				break;
			}
		if(invalue)
			{
			strcat(value, inputFile.space());
			strcat(value, inputFile);
			}
		}
}

EnumMember *parseEnumMember(CommentList &l)
{
	if(debugMode)
		cout << "*** parseEnumMember" << endl;

	static char n[LINESIZE], s[LINESIZE];
	CommentList comments;
	int r = collectEnumMember(comments, n, s);

	if(debugMode)
		cout << "*** collectEnumMember returned " << r << endl;
	if(!r)
		return 0;
	return new EnumMember(n, s, l, comments);
}

void parseEnum(CommentList &comments)
{
	MemberBase *member = 0;
	Enum *current;
	int lineno = inputFile.lineno();

	if(debugMode)
		cout << "*** parseEnum" << endl;
	if(inputFile.nextName() != TTname)
		return;
	current = Enum::theEnum(inputFile);
	current->comments.append(comments);
	comments.zero();
	if(inputFile.nextToken() != '{')
		return;
	for ( ; ; )
		{
		switch(inputFile.nextToken())
			{
			case TTeof:
			case '}':
				finishComment(member, comments);
				return;
			case TTcomment:
				{
				int lastl = (inputFile.lasttoken() == TTcomment) ? 1 : 2;

				if(inputFile.lasttoken() == TTmultiComment
						|| inputFile.lineno() > lineno + lastl)
					finishComment(member, comments, 1);
				comments.append(inputFile);
				lineno = inputFile.lineno();
				continue;
				}
			case TTmultiComment:
				if(inputFile.lasttoken() == TTcomment)
					finishComment(member, comments, 1);
				comments.append(inputFile);
				lineno = inputFile.lineno();
				continue;
			case TTname:
				finishComment(member, comments);
				member = parseEnumMember(comments);
				if(member)
					current->appendMember(member);
				comments.zero();
				lineno = inputFile.lineno();
				continue;
			}
		if(inputFile.lineno() > lineno + 1)
			finishComment(member, comments, 1);
		}
}

void parseClass(CommentList &comments)
{
	Type t = (inputFile.token() == TTstruct) ? Tstruct : Tclass;
	int ip = (t == Tstruct) ? Public : Private;
	MemberBase *member = 0;
	Class *current;
	int lineno = inputFile.lineno();

	if(debugMode)
		cout << "*** parseClass" << endl;
	if(inputFile.nextName() != TTname)
		return;
	current = Class::theClass(inputFile, t);
	current->comments.append(comments);
	comments.zero();
	switch(inputFile.nextToken())
		{
		case ';':		// forward declaration
			return;
		case ':':		// base classes
			parseBases(*current);
		}
	if(inputFile.token() != '{')
		return;
	for ( ; ; )
		{
		switch(inputFile.nextToken())
			{
			case TTeof:
			case '}':
				finishComment(member, comments, 1);
				return;
			case TTcomment:
				{
				int lastl = (inputFile.lasttoken() == TTcomment) ? 1 : 2;

				if(inputFile.lasttoken() == TTmultiComment
						|| inputFile.lineno() > lineno + lastl)
					finishComment(member, comments, 1);
				comments.append(inputFile);
				lineno = inputFile.lineno();
				continue;
				}
			case TTmultiComment:
				if(inputFile.lasttoken() == TTcomment)
					finishComment(member, comments, 1);
				comments.append(inputFile);
				lineno = inputFile.lineno();
				continue;
			case '{':
				inputFile.skipToMatchingBracket();
				break;
			case TTclass:
			case TTstruct:
				finishComment(member, comments);
				parseClass(comments);
				continue;
			case TTenum:
				finishComment(member, comments);
				parseEnum(comments);
				continue;
			case TTprivate:
				finishComment(member, comments, 1);
				ip = Private;
				continue;
			case TTprotected:
				finishComment(member, comments, 1);
				ip = Protect;
				continue;
			case TTpublic:
				finishComment(member, comments, 1);
				ip = Public;
				continue;
			case TTname:
			case '~':
			case TTstatic:
			case TTconst:
			case TTvolatile:
			case TTvirtual:
			case TToperator:
				{
				finishComment(member, comments);

				Member *m = parseMember(ip, comments);

				member = m;
				if(m)
					{
					current->appendMember(m);
					if(m->ispure())
						current->setAbstract();
					}
				comments.zero();
				lineno = inputFile.lineno();
				continue;
				}
			}
		if(inputFile.lineno() > lineno + 1)
			finishComment(member, comments, 1);
		}
}

int readfile(const char *filename)
{
	inputFile.open(filename);
	if(!inputFile)
		return 0;

	int lineno = inputFile.lineno();
	int storageClass = 0;
	CommentList comments;
	MemberBase *member = 0;

	if(debugMode)
		cout << "*** readfile " << filename << endl;
	for ( ; ; )
		{
		switch(inputFile.nextToken())
			{
			case TTignore:
			case TTeof:
				finishComment(member, comments, 1);
				inputFile.close();
				return 1;
			case TTmultiComment:
				if(inputFile.lasttoken() != TTmultiComment)
					finishComment(member, comments, 1);
				comments.append(inputFile);
				lineno = inputFile.lineno();
				continue;
			case TTcomment:
				{
				int lastl = (inputFile.lasttoken() == TTcomment) ? 1 : 2;

				if(inputFile.lasttoken() == TTmultiComment
						|| inputFile.lineno() > lineno + lastl)
					finishComment(member, comments, 1);
				comments.append(inputFile);
				lineno = inputFile.lineno();
				continue;
				}
			case '{':
				inputFile.skipToMatchingBracket();
				break;
			case TTclass:
			case TTstruct:
				finishComment(member, comments);
				parseClass(comments);
				if(debugMode)
					cout << "*** parseClass returned" << endl;
				continue;
			case TTinline:
			case TTconst:
				inputFile.ungetToken();
			case TTstatic:
				storageClass = MTstatic;
				goto doextparse;
			case TTtypedef:
				storageClass = MTtypedef;
				goto doextparse;
			case TTextern:
doextparse:
				finishComment(member, comments);
				member = parseExtern(comments, storageClass);
				if(debugMode)
					cout << "*** parseExtern returned" << endl;
				if(member && member->commentCount())
					member = 0;			// extern commented elsewhere.
				storageClass = 0;
				lineno = inputFile.lineno();
				continue;
			case TTenum:
				finishComment(member, comments);
				parseEnum(comments);
				if(debugMode)
					cout << "*** parseEnum returned" << endl;
				continue;
			}
		if(inputFile.lineno() > lineno + 1)
			finishComment(member, comments, 1);
		}
}

#define PATHLEN 64
int printAll()
{
	GIterator(ClassBase) next(classList);
	ClassBase *s;

	while((s = next()) != 0)
		s->printdoc();
	return 1;
}

void usage()
{
	cout << "Class - extract documentation from class header files" << endl;
	cout << "Usage :-" << endl;
	cout << "class [-options] <file name> ..." << endl;
	cout << "Options are :-" << endl;
	cout << "-f     Keep classes in the same file together" << endl;
	cout << "       Default is to sort in alphabetical order" << endl;
	cout << "-d     Debugging output - more d's = more output" << endl;
	cout << "-p     Print all private members" << endl;
	exit(9);
}

main(int argc, char **argv)
{
	int firstarg, a;
	int sort = 1;
	OutputProcessor p;

	outputProcessor = &p;
	for(firstarg = 1; firstarg < argc && argv[firstarg][0] == '-'; firstarg++)
		{
		for(char *p = argv[firstarg] + 1; *p; p++)
			{
			switch(*p)
				{
				case 'd':
					debugMode++;
					continue;
				case 'f':
					sort = 0;
					continue;
				case 'p':
					printPrivate = 1;
					continue;
				default:
					usage();
				}
			}
		}
	if(firstarg == argc)
		usage();
#ifndef CASE_SENSITIVE_FILENAMES
	for(a = firstarg; a < argc; a++)
		strlwr(argv[a]);
#endif
	for(a = firstarg; a < argc; a++)
		if(!readfile(argv[a]))
			cerr << "Unable to read file " << argv[a] << endl;
	debugMode = 0;
	if(sort)
		classList.sort(alphasort);
	else
		classList.sort(filesort);
	return !printAll();
}

implement(GCollection, char)
implement(GCollection,MemberBase)
implement(GCollection,ClassRef)
implement(GCollection,ClassBase)
implement(GCollection,Class)
implement(GCollection,Enum)
implement(GCollection,Extern)


