#ifdef __ZTC__
#include <generic.hpp>
#else
#include <generic.h>
#ifndef name2
#define name2(z, y) z ## y
#endif
#ifndef name3
#define name3(z, y, x) z ## y ## x
#endif
#endif

// $Revision: 1.8 $

/*
This base Collection class will hold any object, as the implementation is
a dynamic array of void pointers. The generic class macros are used to
provide type-safe Collections of any particular kind of object. A useful
variety of methods are provided, together with a copy constructor and
assignment operator, so you can assign Collections to one another without
problems. 
*/
class Collection {
public:
	Collection(int hintSize = 0);
	Collection(const Collection& c);
	virtual ~Collection();

	Collection& operator= (const Collection& c);
	Collection& operator+ (const Collection &c);
	Collection& operator+ (void *v) { append(v); return *this; }

	void append(void *c);	  // Add item to end of list.

	int size() const				// Number of elements in the list.
		{ return uelements; }
	void *first() const				// Return first item on list.
		{ return (uelements ? list[0] : 0); }
	void *last() const				// Return last item on list.
		{ return (uelements ? list[uelements - 1] : 0); }
	void *getlast() { return (uelements) ? list[--uelements] : 0; }
	// Remove and return the last item on the list.

	void setsize(int n)			// Make big enough for at least n elements.
		{ if(n > nelements) expand(nelements - n); }
	int expand(int n = 1);		// Make the array n elements bigger.

	virtual int compare(const void *v1, const void *v2) const;
	// Compare two elements using the current sort function.
	void resort();

protected:
friend class Iterator;
	void **list;					// Pointer to contents.
	int nelements;					// Number of elements in array.
	int uelements;					// Number of elements in use.
};

/*
This class provides an iterator to step over the elements of a
Collection one at a time. The class allows traversal in either
direction, and maintains a default direction set up in the
constructor. The Iterator can even be started at a particular
element in the Collection, rather than at the beginning (or end if
traversal is backwards). 
*/
class Iterator {
	const Collection &list;
	int i;								// current element no
public:
	Iterator(const Collection& l);

	void reset() { i = -1; }			// Move pointer to start of list.
	void *next();						// Next item on list
	void *operator() () { return next(); };		// Syntactic sugar for next
};

#define GCollection(type) name2(GCollection,type)

/*
A generic Collection to hold items of a particular type. See
Collection for method details. Note that the sort method will
now require a function which takes two pointers to items of your
type, rather than void *. To declare a Collection class to
hold Widgets (say), all you need to do is ...

declare(GCollection,Widget); 

in your header (.hpp) file. This will provide the definition of a
Collection class, called GCollection(Widget), including all
the inline functions. Then, in your .cpp file, do ... 

implement(GCollection,Widget) 

(Note lack of semicolon), which will implement the non- inline
functions. 
*/
#define GCollectiondeclare(type)\
class GCollection(type): public Collection {\
	int (*compfn)(const type *,const type *);\
public:\
	GCollection(type)(int hintSize = 0);\
	GCollection(type)(const GCollection(type)& c);\
	GCollection(type)& operator= (const GCollection(type)& c) \
		{ Collection::operator=(c); compfn = c.compfn; return *this; } \
	GCollection(type)& operator+ (const GCollection(type) &c) \
		{ Collection::operator+(c); return *this; } \
	GCollection(type)& operator+ (type *c) { append(c); return *this; }\
	void append(type *item) { Collection::append(item); }\
	type *first() const { return (type *)Collection::first(); }\
	type *last() const { return (type *)Collection::last(); }\
	type *getlast() { return (type *)Collection::getlast(); }\
	void sort(int (*compare)(const type *,const type *))\
		{ compfn = compare; resort(); }\
	virtual int compare(const void *v1, const void *v2) const;\
}

#define GCollectionimplement(type)\
GCollection(type)::GCollection(type)(int hintSize)\
: Collection(hintSize), compfn(0) {}\
GCollection(type)::GCollection(type)(const GCollection(type)& c)\
: Collection(c), compfn(0) {} \
int GCollection(type)::compare(const void *v1, const void *v2) const\
{\
	return compfn ? compfn((type *)v1,(type *)v2) : 0;\
}

#define GIterator(type) name2(GIterator,type)

/*
An Iterator which iterates over a generic Collection -
GCollection(type). To declare an Iterator class for a
GCollection(Widget) (say), all you need to do is ...

declare(GIterator,Widget); 

This will create an Iterator class, called GIterator(Widget).
Note that all the functions are inline, so no further implementation
is needed. 
*/
#define GIteratordeclare(type)\
class GIterator(type) : public Iterator {\
public:\
	GIterator(type)(const GCollection(type)& s) : Iterator(s) {}\
	type *next() { return (type *)Iterator::next(); }\
	type *operator()() { return next(); }\
}

#define LINESIZE 1024
#define BIT(n) (1 << (n))

// Return values for LineFile::nextToken
// May also return the ASCII value of a punctuation character.
enum TokenType {
	TTbegins = -100,
	TTname = TTbegins,
	TTnumber,
	TTstring,
	TTarrayref,
	TTchar,
	TTcomment,
	TTmultiComment,
	TTignore,
	TTarglist,
	TTkeywords,
	TTclass = TTkeywords,
	TTstruct,
	TTenum,
	TTprivate,
	TTpublic,
	TTprotected,
	TTstatic,
	TTextern,
	TTtypedef,
	TTconst,
	TTvolatile,
	TTvirtual,
	TTinline,
	TToperator,
	TTeof = -1
	};

// This class deals with an input file, and parsing it into tokens
class LineFile {
public:
	LineFile(const char *fname = "");
	~LineFile();
	int open(const char *fname);
	int isopen() { return (*fname && f != 0); }
	operator int() { return isopen(); }
	operator!() { return !isopen(); }
	// So LineFile objects can be used in an if statement
	void close();
	const char *filename() const { return fname; }
	operator const char *() const { return tok; }
	// Returns the current token
	int token() const { return tokt; }
	// Returns the type of the current token
	const char *last() const { return lasttok; }
	// Returns the previous token
	int lasttoken() const { return lasttokt; }
	// Returns the type of the previous token
	int lineno() const { return lno; }
	const char *space() const { return lastspace; }
	// Returns the sequence of whitespace which preceeded the current token
	int inComment() const { return inMultiComment; }
	// True if currently in a multi-line comment
	int nextToken();
	// Get the next token, and return its type
	int nextName();
	// Get the next sequence of tokens making up a C++ name
	int getRestOfName();
	// Having got a token which might be a template/generic name,
	// add the template arguments to the end of the name token.
	void ungetToken();
	// Put a token back on the input stream
	int operator ()() { return nextToken(); }
	// Syntactic sugar for nextToken
	int operator==(const char *string) const 
	// Compare token with supplied string
		{ return (strcmp(*this,string) == 0); }
	int operator!=(const char *string) const 
	// Compare token with supplied string
		{ return (strcmp(*this,string) != 0); }
	void skipToMatchingBracket(char bracket = '{');
	// Skip input until matching bracket ({, (, < or [)
private:
	ifstream f;
	const char *fname;
	char *cpos;
	char lasttok[LINESIZE+1];
	char tok[LINESIZE+1];
	char holdtok[LINESIZE+1];
	char lastspace[LINESIZE+1];
	char lookahead[10];
	char *ungot;
	unsigned lno;
	int inMultiComment;
	int tokt;
	int lasttokt;
	int holdtokt;
	int get();
	void unget(int c);
	void skipstring(char delim = '"');
	void skipbracket(char bracket = '(');
	void skipcomment(char *copy = 0);
	int skipmultiComment(char *copy = 0);
	int skipspace();
	int getnext();
	};

// Dummy structure - output to an OutputProcessor at the end of each line.
struct LineEnd {
	} endLine;

// A very simple post-processor for the output. 
// You can create derived classes for printer-specific output if you wish.
class OutputProcessor {
protected:
	enum { lineBufferSize = 1024 };
	char lineBuffer[lineBufferSize];
	int headingLevel;
	int indentLevel;
public:
	OutputProcessor();
	virtual ~OutputProcessor();
	void underline(char linechar = '-');			// Set underline character.
	void heading(int level);		// Set heading character.
	void indent(int n);				// Set indent amount.
	virtual void includeFile(const char *filename);	// Print include file name.
	virtual void endLine();			// Print the current line from lineBuffer.
	OutputProcessor &operator<<(const char *string);
	OutputProcessor &operator<<(const LineEnd l);
	};

// Possible values for 'level' in OutputProcessor::underline.
enum HeadingLevel {
	HLfirst,
	HLclass = HLfirst, 
	HLstruct, 
	HLenum, 
	HLstatic, 
	HLtypedef,
	HLextern,
	HLbaseClasses, 
	HLderivedClasses,
	HLpublics, 
	HLprotecteds, 
	HLprivates, 
	HLinheritedPublics, 
	HLinheritedProtecteds,
	HLmember,
	HLinheritClass,
	HLenumValues,
	HLlast,
	HLnone
	};

class Class;
typedef void (Class::*ClassFunc)(int) const/*!*/;
// Used for inheritance, and class methods and variables
enum Protection { 
	Public, 
	Protect, 
	Private, 
	Virtual = 4 	// Use a spare bit for virtual inheritance
	};

// Used for class methods/variables, and also extern & static 
// functions & variables
enum MemberType { 
	MTpublic = 0,
	MTprotected = BIT(0),
	MTprivate = BIT(1),
	MTprotection = (MTprotected|MTprivate),	// Mask for protection bits
	MTinline = BIT(2), 
	MTvirtual = BIT(3), 
	MTpure = BIT(4),
	MTconst = BIT(5),
	MTdestructor = BIT(6),
	MTstatic = BIT(7),
	MTtypedef = BIT(8),
	MTexternC = BIT(9)
	};

declare(GCollection, char);
declare(GIterator, char);

/*
A list of member names, so that members that have been overridden
in derived classes are not printed twice when recursing through base
classes.
*/
class NameList {
	GCollection(char) list;
public:
	NameList();
	~NameList();
	void append(const char *m);
	void append(NameList &n);
	int contains(const char *);
	void zero();		// Empty the list, deleting the contents
	};

/*
Contains the last complete block of comments for later output as
the class description.
*/
class CommentList {
	GCollection(char) list;
public:
	CommentList();
	~CommentList();
	void append(const char *m);
	void append(CommentList &l);
	void zero();
	void print(int indent = 0) const;
	int size() const { return list.size(); }
	};

// Holds a class/enum/whatever member, and the comments associated with it
class MemberBase {
protected:
	char *name;
	CommentList preamble;		// Comments before the declaration
	CommentList comments;		// Comments after the declaration
public:
	MemberBase(const char *membername, CommentList &before, CommentList &after);
	virtual ~MemberBase();
	const char *namepart() const { return name; }
	// Returns the plain unadorned member name.
	// Does not include ~ of destructors, return type, parameters etc.
	void append(CommentList &c) { comments.append(c); }
	// Add more comments found after the declaration
	int commentCount() const { return comments.size(); }
	// The number of lines of comments after the declaration
	virtual void print() const = 0;
	};

// Holds a class member
class Member : public MemberBase {
	char *sig;
	int type;
public:
	Member(int type, const char *n, const char *s, 
		CommentList &p, CommentList &c);
	virtual ~Member();
	int operator==(Member &m) { return !strcmp(sig, m.sig); }
	// Are the names and signatures the same ?
	const char *signature() const { return sig; }
	int isprotected() const { return (type & MTprotected); }
	int isprivate() const { return (type & MTprivate); }
	int protection() const { return (type & MTprotection); }
	int isinline() const { return (type & MTinline); }
	int isvirtual() const { return (type & MTvirtual); }
	int ispure() const { return (type & MTpure); }
	int isconst() const { return (type & MTconst); }
	int isdestructor() const { return (type & MTdestructor); }
	int isstatic() const { return (type & MTstatic); }
	int istypedef() const { return (type & MTtypedef); }
	const char *storageClass() const 
		// For printing - returns "", "static " or "typedef " or ""
		{ return (isstatic() ? "static " : (istypedef() ? "typedef " : "")); }
	virtual void print() const;
	};

// Holds an enum member
class EnumMember : public MemberBase {
	char *val;
public:
	EnumMember(const char *n, const char *v, CommentList &p, CommentList &c);
	virtual ~EnumMember();
	int operator==(EnumMember &m) 
	// Are the names the same ?
		{ return !strcmp(name, m.name); }
	const char *value() const { return val; }
	virtual void print() const;
	};

declare(GCollection,MemberBase);
declare(GIterator,MemberBase);

// A reference to a sub or super class.
class ClassRef {
	Class *info;
	int prot;
public:
	ClassRef(Class *c, int p = Public);
	int protection() const { return (prot & ~Virtual); }
	int isvirtual() const { return (prot & Virtual); }
	Class *theClass() const 
	// Return the class referred to
		{ return info; }
	operator Class *() const { return info; }
	// User defined conversion to the class referred to
	friend OutputProcessor& operator<<(OutputProcessor &o, const ClassRef &r);
	// Prints name and protection
	void print(int indent = 0,ClassFunc func = 0) const;
	// Recursive printing function
	};

declare(GCollection,ClassRef);
declare(GIterator,ClassRef);

// Distinguish classes and structs
enum Type { 
	Tclass, 
	Tstruct, 
	Tabstract = BIT(1)	// Spare bit used to indicate abstract class
	};

// Translate enum Type into text string
const char *const typenames[] = { 
	"class", 
	"structure", 
	"abstract class", 
	"abstract structure" 
	};

// Base class for classes, structs, enums, externs & statics
class ClassBase {
protected:
	ClassBase *next;
	char *cname;
	const char *fname;
	int lno;
	ClassBase(const char *name);
public:
	virtual ~ClassBase();
	const char *name() const { return cname; }
	const char *filename() const { return fname; }
	friend OutputProcessor& operator<<(OutputProcessor &o, const ClassBase &b)
		{ o << b.name(); return o; }
	// Prints name
	virtual void printdoc() const = 0;
	int lineno() const { return lno; }
	};

declare(GCollection,ClassBase);
declare(GIterator,ClassBase);
GCollection(ClassBase) classList, emptyList;

// Base class for classes, structures and enums
class BaseWithMembers : public ClassBase {
protected:
	GCollection(MemberBase) members;
	BaseWithMembers(const char *name);
public:
	CommentList comments;
	virtual ~BaseWithMembers();
	const GCollection(MemberBase) &classMembers() const { return members; }
	void appendMember(MemberBase *m);
	};

// The main information about a class.
// Records base classes, derived classes, members and comment text.
class Class : public BaseWithMembers {
	GCollection(ClassRef) bases;
	GCollection(ClassRef) derived;
	Type t;
	Class(const char *name, Type type = Tclass);
	void printSomeMembers(NameList &names, int type) const;
	static NameList mynames;
	static int doingBases;
public:
	virtual ~Class();
	Type type() const { return (Type)(t & ~Tabstract); }
	int isabstract() const { return (t & Tabstract); }
	void setAbstract() { t = (Type)(t | Tabstract); }
	const char *typeString() const { return typenames[t]; }
	operator const GCollection(ClassRef) &() const { return bases; }
	const GCollection(ClassRef) &baseClasses() const { return bases; }
	const GCollection(ClassRef) &derivedClasses() const { return derived; }
	void appendBase(Class *c, int p);
	static Class *theClass(const char *name, Type t = Tclass);
	// Look up the name in the list of all classes of type t
	// Return the Class object (creating a new one if necessary)
	void print(int indent = 0,ClassFunc func = 0) const;
	void printbases(int indent = 0) const;
	void printderived(int indent = 0) const;
	void printmembers(NameList &names) const;
	virtual void printdoc() const;
	};

declare(GCollection,Class);
declare(GIterator,Class);
GCollection(Class) classOnlyList;

// The main information about an enum
// Records members and comment text.
class Enum : public BaseWithMembers {
	Enum(const char *name);
public:
	virtual ~Enum();
	static Enum *theEnum(const char *name);
	void printmembers() const;
	virtual void printdoc() const;
	};

declare(GCollection,Enum);
declare(GIterator,Enum);
GCollection(Enum) enumOnlyList;

// The information about an extern
class Extern : public ClassBase, public MemberBase {
	int type;
	Extern(int type, const char *n, const char *v, 
		CommentList &p, CommentList &c);
public:
	virtual ~Extern();
	static Extern *theExtern(int type, const char *name, 
		const char *v, CommentList &p, CommentList &c);
	const char *signature() const { return MemberBase::namepart(); }
	int isinline() const { return (type & MTinline); }
	int isconst() const { return (type & MTconst); }
	int isstatic() const { return (type & MTstatic); }
	int isexternC() const { return (type & MTexternC); }
	int istypedef() const { return (type & MTtypedef); }
	const char *storageClass() const
	// Returns "static ", "typedef " or "extern "
		{ return (isstatic() ? "static " : (istypedef() ? "typedef " : "extern ")); }
	virtual void printdoc() const;
	virtual void print() const;
	};

declare(GCollection,Extern);
declare(GIterator,Extern);
GCollection(Extern) externOnlyList;


