// Event handling
#ifndef EVENT_HPP
#define EVENT_HPP
#ifndef RECT_HPP
#include "rect.hpp"
#endif
#ifndef COLLECT_HPP
#include "collect.hpp"
#endif
#ifndef PROTECT_HPP
#include "protect.hpp"
#endif

class Window;

// An abstract class for the basic external event on the EventQueue. 
class Event {
public:
	virtual ~Event();
	// Force virtual destructor.
	virtual int despatch(Window *control) = 0;
	// Decide which VObjects in the Application are to handle
	// the event. Calls handle() for each one in turn until handle
	// returns non-zero. May call handle with special value of 0,
	// meaning send the event to the Application.
	// This method will be overridden in derived classes - e.g. a
	// KeyboardEvent despatches itself to the keyboard focus, a MouseEvent
	// despatches itself to the mouse focus, or to the object in which the
	// event happened. 
	virtual int handle(VObject *o) = 0;
	// Send the Event to the specified VObject. Return 0 if the
	// VObject was not interested. May be called with a special
	// value of 0, meaning send the Event to the Application
	};

/*
An abstract Event which is despatched to the keyboard focus and its 
containers. If they are not interested, it is sent to the 
Application.
*/
class FocusEvent : public Event {
public:
	virtual int despatch(Window *control);
	virtual int handle(VObject *o) = 0;
	};

/*
An Event with a position, which is despatched to the top Window
which contains the position. 
*/
class PositionEvent : public Event {
protected:
	Point p;
public:
	PositionEvent(const Point &here = origin);
	const Point &at() const { return p; }
	// Position where the event happened.
	virtual int despatch(Window *control);
	virtual int handle(VObject *o) = 0;
	};

/*
An external source of Events - e.g. a mouse, or the keyboard. A
derived class exists for each possible source of Events. You can
easily add your own EventSources by deriving from this class. The
toolkit already contains internal static EventSources for the mouse,
keyboard and control-c handler. These sources are automatically
added to the EventQueue when it is created. You can create your own
event sources and add them to the queue if you wish. 
*/
class EventSource : public NamedProtected {
	virtual Event *v_poll() = 0;
	// Determine if there is an Event waiting at this device. 
	// Return 0 if there is no event waiting, or a pointer to the Event. Most 
	// implementations return a pointer to an internal static.
	virtual int v_enable();
	// Setup code (e.g. hooking interrupts) for the EventSource. 
	// Will be called when the EventQueue is enabled (after the Screen 
	// object is initialised). It may also be called again (after a call 
	// to disable) - e.g. after shelling to DOS. 
	// Return 1 on success, 0 on failure. 
	// Will only be called if the EventSource is not currently enabled. 
	virtual void v_disable();
	// The method you should override to contain any tidy up code (e.g.
	// restoring interrupts) for the EventSource. It will be called when
	// the EventQueue is disabled (before the Screen object is restored).
	// It may also be called during the execution of the program, e.g.
	// before shelling to DOS. 
	// Will only be called if the EventSource is currently enabled. 
	int enabledOK;
public:
	EventSource(const char *name);
	virtual ~EventSource();
	int enabled() const { return enabledOK; }
	Event *poll() { VALIDATE(this); return enabled() ? v_poll() : 0; }
	// This method actually calls the private virtual v_poll(), which is
	// the method to override to actually determine if there is an 
	// Event waiting at this device. It should return 0 if there is 
	// no event waiting, or a pointer to the Event. Most 
	// implementations return a pointer to an internal static.
	int enable() { VALIDATE(this); return enabled() ? 1 : (enabledOK = v_enable()); }
	// This method actually calls the private virtual v_enable(), which is
	// the method you should override to contain any setup code (e.g.
	// hooking interrupts) for the EventSource. It will be called when the
	// EventQueue is enabled (after the Screen object is initialised). It
	// may also be called again (after a call to disable) - e.g. after
	// shelling to DOS. Return 1 on success, 0 on failure. The v_enable
	// virtual method will only be called if the EventSource is not
	// currently enabled. 
	void disable() { VALIDATE(this); if(enabled()) { v_disable(); enabledOK = 0; } }
	// This method actually calls the private virtual v_disable(), which is
	// the method you should override to contain any tidy up code (e.g.
	// restoring interrupts) for the EventSource. It will be called when
	// the EventQueue is disabled (before the Screen object is restored).
	// It may also be called during the execution of the program, e.g.
	// before shelling to DOS. The v_disable virtual method will only be
	// called if the EventSource is currently enabled. 
	};

class MouseEvent;
typedef int (VObject::*mouseEventFunc)(MouseEvent &);

// The possible mouse buttons which may be pressed.
enum MouseButtons
	{
	MBnone = 0,
	MBleft = BIT(0),
	MBright = BIT(1),
	MBmiddle = BIT(2),
	MBinspect = BIT(3)
	};

/*
A MouseEvent is placed in the event queue whenever the mouse is
moved, or a button is depressed or released. 
*/
class MouseEvent : public PositionEvent {
private:
	mouseEventFunc func;
	const char *namestring;
	MouseButtons btns;
friend class Mouse;
public:
	MouseEvent(mouseEventFunc handler = 0,const char *name = "",
		const Point &here = origin, MouseButtons buttons = MBnone);
	// The constructor is passed a pointer to a VObject method which will
	// be called whenever the MouseEvent's handle method is called. 
	virtual int handle(VObject *o);
	// Pass the MouseEvent to the Vobjects handle method. This method may
	// call the MouseEvent's callHandler method to re-despatch the
	// MouseEvent to the intended VObject method. The default handle method
	// first calls the mouseenter or mouseexit methods if necessary, then,
	// if they do not return 1, it calls back the callHandler method as
	// described. 
	int callHandler(VObject *o) { return (func ? (o->*func)(*this) : 0); }
	// Pass the Event to the VObject method pointer given in the constructor.
	const char *name() const { return namestring; }
	MouseButtons buttons() const { return btns; }
	// Which buttons are pressed. See the MouseButtons enum.
static int doubleClickTime;
	// This is set to the time (in hundredths of a second) which must
	// elapse between two clicks for them not to be treated as a double
	// click. The default value is 50 (half a second). 
	};

// Event representing a mouse movement.
class MouseMovement : public MouseEvent {
public:
	MouseMovement(const Point &here);
	MouseMovement();
	virtual int handle(VObject *o);
	};

#ifndef NINSPECT
// Event representing a mouse inspect-click (Alt-left click).
class MouseInspect : public MouseEvent {
public:
	MouseInspect(const Point &here = origin);
	virtual int handle(VObject *o);
	};
#endif

// Event representing a key depression.
class KeyboardEvent : public FocusEvent {
friend class Keyboard;
public:
	KeyboardEvent();
	virtual int despatch(Window *control);
	// Each character of keyboard input is passed to the FocusEvent
	// despatch method, which tries to pass it to the keyboard focus of the
	// Application, or its containing VObjects. If none of these are
	// interested, the various hotkey methods are tried in turn. 
	virtual int handle(VObject *o);
	// Pass the event to the VObject kbdevent method.

	unsigned value() const { return val; }
	// Return the ASCII value of the key pressed, or one of the values from
	// the Keys enum. 
	unsigned shifts() const { return shstate; }
	// Return the shift state (as returned by the BIOS). This state might
	// be meaninless under Unix. 

private:
	unsigned val;
	unsigned shstate;
	};

/*
The event generated whenever an interrupt signal occurs (Ctrl-Break
under DOS). The default Application traps this event, and puts up a
confirmation box before exiting the program. 
*/
class InterruptEvent : public FocusEvent {
public:
	virtual int handle(VObject *o);
	// The Event is passed to the VObject's interruptevent method. 
	};

// DOS critical error codes.
enum CEerrorCode {
	CEwriteProtected,
	CEnoSuchDrive,
	CEnotReady,
	CEbadCommand,
	CEdataError,
	CEbadRequest,
	CEseekError,
	CEmedia,
	CEsectorNotFound,
	CEpaperOut,
	CEwriteError,
	CEreadError,
	CEgeneralError
	};

/*
The Event generated when a DOS critical error occurrs while a
CriticalErrorHandler is in existence. 
*/
class CriticalErrorEvent : public FocusEvent {
friend class CriticalErrorSource;
public:
	virtual int handle(VObject *o);
	// Pass the Event to the VObject's criterr method.
	int drive() const { return d; }
	// The DOS drive letter which caused the error, or '?' if no drive
	// applies. 
	int ax() const { return a; }
	// The value from the AX register when the DOS critical error handler
	// was called. See your DOS programmers manual for details. 
	CEerrorCode errcode() const { return e; }
	// The error code.
	const char *description() const;
	// A standard 2-line description of the error. Useful if you want to
	// pop up an alert box describing the error to the user. 

private:
	int d, a;
	CEerrorCode e;
	};

/*
A class to trap DOS critical errors. 
All you need to do is to declare a CriticalErrorHandler in a
function or block. Critical errors will be trapped from the
declaration to the end of the enclosing block. This is one of the
few objects in the toolkit that is declared directly, rather than
being allocated via new. 
*/
class CriticalErrorHandler {
public:
	CriticalErrorHandler();
	~CriticalErrorHandler();
	};

class Command;

/*
An Event which will execute the specified Command when it is 
despatched. Used as a method of passing a Command through the 
event queue.
*/
class CommandEvent : public Event {
	Command *cmd;
public:
	CommandEvent(Command *command);
	virtual int despatch(Window *control);
	virtual int handle(VObject *o);
	};

declare(GCollection,EventSource);
declare(GIterator,EventSource);

/*
The main event queue for the system. Only one EventQueue is created,
called eventQueue. This must be created in your main program, before
the Application. It is enabled when the Application starts up, and
polled by the Application. You can add new EventSources to those
polled by the queue. 
*/
class EventQueue : private GCollection(EventSource), public Protected {
static int cons;
public:
	EventQueue();
	~EventQueue();
static int constructed() { return cons; }
	void enable();
	// Enable all EventSources (unless they are already enabled). Sets the
	// enabled flag on each EventSource successfully enabled. 
	void disable();
	// Disable all EventSources (unless they are already disabled). Clears
	// the enabled flag on each EventSource disabled. 
	int add(EventSource *item);
	// Add a new EventSource. The EventSource should normally be enabled
	// when adding, but this is not necessary if you intend to enable it
	// later. The enabled flag determines whether the EventSource is polled.
	// Returns -1 on failure. 
	int remove(EventSource *item);
	Event *operator() ();
	// Returns the next Event from the EventQueue. Returns 0 if no event is
	// waiting. 
	};

extern EventQueue eventQueue;
#endif
