// A more realistic application.  Read in and display a list of
// telephone numbers.  Each line of the input file should contain
// a one-letter cost centre code, the name, and the telephone number.
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
// The includes for the various toolkit components
#include "panel.hpp"	/* A Panel contains the main menu and client area */
#include "applic.hpp"	/* The application */
#include "menu.hpp"		/* The menu */
#include "list.hpp"		/* The list of numbers */
#include "dialog.hpp"	/* For the file open dialog */
#include "button.hpp"	/* The buttons */
#include "fileedit.hpp" /* For the file open dialog */
#include "text.hpp"		/* Text for prompt */
#include "event.hpp"	/* To make the event queue */

#define MAXLINE 256		/* Maximum input line */
#define MAXSEARCH 40	/* Maximum search string length */

// A class to hold phone numbers
// Derived from ListElement, which displays its name in a List
class Number : public ListElement {
private:
	char *k;	// The whole thing - "<cost code> <name> <number>"
	char *n;	// Points at the number part of k
	// initialise the variables given an input line
	// We keep this separate from the constructor, because we intend
	// to introduce more constructors as we expand the program.
	void build(const char *inpline);

public:
	// Construct from a file input line
	Number(const char *inpline);
	// Tidy up allocated memory
	~Number();

	// Access methods.
	const char *key() const { return k + 2; }
	const char *number() const { return n; }
	char code() const { return *k; }
	};

// We are going to hold Numbers in a List, so we need a Collection
// to put them in.
declare(LCollection,Number);	// Something to hold a list of numbers
declare(LIterator,Number);		// An Iterator for the list
implement(LCollection,Number)

// We will allow several different types of search action
enum ScanMode {
	SMfind,		// Find the first key beginning with search string
	SMscan,		// Scan for the first entry containg the string anywhere
	SMnext,		// Repeat last search forwards
	SMprevious, // Repeat last search backwards
	SMcount		// The number of values for this enum (for array sizes)
	};

// Now lets derive a List class to hold the Numbers.  We could use
// the basic List class on its own, but deriving a class demonstrates
// how this should be done, and will make the program more extensible.
class PhoneList : public List {
private:
	LCollection(Number) directory;	// Collection of numbers from the file
	ScanMode scanmode;				// How we are searching this list
	char fname[OSPathLen+1];		// The file name

public:
	// Constructor
	PhoneList(const char *filename = "PHONE.DIR");
	// Access method - return the currently selected number (or 0 if none)
	Number *selected() const { return directory[line()]; }
	// Read the List from a file. Use current file name if none supplied.
	int read(const char *filename = 0);
	// Find a string in the list
	int find(const char *string, ScanMode s = SMfind);
	// Force a complete re-display of the List
	void dirty() { changeCollection(); List::dirty(); }
	int openFile(const char *);		// Open file dialog
	int findNumber(const char *button);
	};

declare(GCommand,PhoneList);
implement(GCommand,PhoneList)

// Now the constructor
Number::Number(const char *inpline)
: ListElement("")		// We will set the name to the data later
{
	while(isspace(*inpline))		// Skip leading spaces
		inpline++;

	char c = isalpha(*inpline) ? *inpline++ : ' ';	// Get the code

	while(isspace(*inpline))		// Skip any spaces between code and name
		inpline++;
	build(form("%c %s",c,inpline)); // The real "meat" of the constructor
}

// Destructor - delete the allocated memory
Number::~Number()
{
	newname("");		// we are about to delete name !
	delete k;
}

// The real construction work !
void Number::build(const char *inpline)
{
	int l = strlen(inpline);	// Length (including terminating newline)

	n = k = new char[l];		// Allocate the memory
	if(!k)
		return;					// failed !
	strncpy(k,inpline,l);		// Copy excluding terminating newline
	k[l - 1] = 0;				// Ensure null teminated
	for(n = k; *n && !isdigit(*n); n++)
		;						// Search along for the number
	newname(k);					// Tell the ListElement what to display
}

// Constructor
PhoneList::PhoneList(const char *filename)
// Pass the Collection, name, and border style to List
// The border will be a single line, the list will have a title,
// and the user will be allowed to Tab to the list.
: List(directory,strcpy(fname,filename),BorderFlag(BFsingle|BFtitle|BFtabstop))
{
	read();			// Read the data from the file
}

// Read the list from a file
int PhoneList::read(const char *filename)
{
	if(!filename)
		filename = fname;			// Use the default file name

	FILE *fd = fopen(filename,"r");

	if(fd == 0)
		return 0;					// Open failed - file not found ?
	directory.deleteall();			// Delete any existing contents

	static char inpline[MAXLINE+1]; // To hold the input lines.

	while(fgets(inpline,MAXLINE,fd))// Read lines from the file until EOF
		{
		if(inpline[0] == '\n' || inpline[0] == 0)
			continue;				// Skip blank lines
		directory.add(new Number(inpline)); // Add to Collection
		}
	fclose(fd);
	if(filename != fname)
		strcpy(fname,filename);		// Remember file name
	newname(fname);					// and set the title to it.
	dirty();						// Force a redraw
	return 1;
}

// A utility function to find if a string starts with a pattern.
// Case is ignored in the search.
int strstart(const char *string, const char *pattern)
{
	for( ; *pattern; string++, pattern++)
		if(tolower(*string) != tolower(*pattern))
			return 0;		// Not found
	return 1;				// Found
}

// A utility function to find if a string contains a pattern anywhere in it.
int scan(const char *string, const char *pattern)
{
	int l = strlen(string) - strlen(pattern) + 1;

	for( ; l-- > 0; string++)
		if(tolower(*string) == tolower(*pattern)	// optimise a bit
			&& strstart(string,pattern))
				return 1;
	return 0;
}

// Search the PhoneList for a pattern
int PhoneList::find(const char *pattern, ScanMode s)
{
	Number *c = 0;			// Where to start the search
	int backwards = 0;		// Direction of search

	switch(s)
		{
	case SMfind:			// Search forwards
	case SMscan:			// Scan forwards
		scanmode = s;		// Remember search type for next time
		break;
	case SMprevious:
		backwards = 1;		// Search backwards
		/* drop through */
	case SMnext:			// Search forwards
		c = selected();		// From currently selected item
		}

	// We want to travrse the Collection, so declare an Iterator
	// Note that a simple forward search only requires
	// LIterator(Number) next(directory);
	LIterator(Number) next(directory,c,backwards);

	while((c = next()) != 0)			// Traverse the Collection
		{
		if(strstart(c->key(),pattern)	// Match or scan for pattern
										// depending on search mode
			|| (scanmode == SMscan && scan(c->key(),pattern)))
			{							// Found - move selection to element
			select(directory.elementNo(c));
			return 1;
			}
		}
	return 0;							// Not Found
}

// A standard function, called whenever the "Open"
// Option in the "File" Menu is selected.
int PhoneList::openFile(const char *)
{
	// We want to input a file name, so create a FileEdit
	FileEdit *f = new FileEdit(fname,"Number file");

	// A ModalDialog interacts with the user, then returns
	ModalDialog *dlg = new ModalDialog(
		// Lay out elements in a vertical Cluster
		new VCluster(ObjectCollection(2)
			+ f			 // The FileEdit
		// Now a horizontal Cluster of Buttons
			+ new HCluster(ObjectCollection(2)
		// A CloseButton which updates the FileEdit with the user's input
				+ new CloseEditButton(f)
		// and a CancelButton, in case the user changes their mind
				+ new CancelButton
				)
			),
		// The title of the Dialog
		"Open File");
	if(!dlg->enter())		// Interact with the user
		read();		// If "O.K." pressed, read the file
	return 1;
}

char searchstring[MAXSEARCH+1];			// The pattern to find
const char *buttonNames[SMcount] =		// The different search buttons
	{ "F3=Find", "F4=Scan", "F5=Next", "F6=Previous" };

// Now, we need a function (of a standard type) to be called when the
// user presses a search Button. This function will be called automatically
// when one of the Buttons is pressed. Note that we have only one function
// here, and we look at the name of the Button which was pressed to
// find out which search strategy to use.  Aside from the advantage that
// we only have to write one function, it will not have to change if
// we add new types of search Button.
int PhoneList::findNumber(const char *button)
{
	for(int j = 0; j < SMcount; j++)
		if(!strcmp(button,buttonNames[j]))
			return find(searchstring,(ScanMode) j);
	return 0;
}

// A standard type function for the "Exit" Option on the "File" Menu
int exitProg(const char *)
{
	exit(0);
	return 1;
}

EventQueue eventQueue;	// We always have an EventQueue
Application example;	// We always have an Application

main()
{
	PhoneList *phonelist = new PhoneList();
	// Now an Edit to input the search string
	TextEdit *search = new TextEdit(searchstring,MAXSEARCH,"Search string");
	// and the first (default) "Find" button.  It must update searchstring
	// with the user's input, and it must call findNumber whenever it
	// is pressed.
	ApplyEditButton *find = new ApplyEditButton(search,
		buttonNames[SMfind],KF3,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist));

	find->makedefault();		// Make this the default Button
								// (selected when the user presses Return)
	// and the Panel
	Panel *main = new Panel(
	// Panels always have a MenuBar
		new MenuBar(KF10,			// This is the SAA standard hot-key
			OCollection(MenuButton)(2)
	// The first Menu in the MenuBar
			+ new MenuButton("~File",KAltF,LCollection(Option)(2)
				+ new Option("~Open",new GCommand(PhoneList)(&PhoneList::openFile,phonelist))
				+ new Option("~Exit",exitProg,KAltF4)
				)
	// Allow context-sensitive help
			+ DefaultHelpMenu
			),
	// Now the client area, which is arranged vertically
		new VCluster(ObjectCollection(3)
	// the search string prompt, and field
			+ new HCluster(ObjectCollection(2)
				+ new FixedText("Search for ")
				+ search
				)
	// The search Buttons
			+ new HCluster(ObjectCollection(4)
				+ find	 // we defined this one earlier
	// Here are the others
				+ new ApplyEditButton(search,
					buttonNames[SMscan],KF4,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist))
				+ new ApplyEditButton(search,
					buttonNames[SMnext],KF5,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist))
				+ new ApplyEditButton(search,
					buttonNames[SMprevious],KF6,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist)),
				"Search list",AFHexpand,BorderFlag(BFsingle|BFtabstop))
	// The PhoneList itself
			+ phonelist
			),
	// Panel name
		"Phone numbers");
	main->show();		// Place the Panel on the screen
	// Now open the help file. This is all you have to do to add context
	// sensitive help to an application.  Of course, you have to
	// write the actual help text - see "PHONE.SRC" for details.
	Application::openHelpFile("phone");
	example.enter();	// Enter the main message loop
	return 0;
}
