// Now we provide the ability to edit and print the list
// and we trap DOS critical errors.
// Changes from example3 are marked with "//**"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
// The includes for the various toolkit components
#include "panel.hpp"	/* A Panel contains the main menu and client area */
#include "applic.hpp"	/* The application */
#include "menu.hpp"		/* The menu */
#include "list.hpp"		/* The list of numbers */
#include "dialog.hpp"	/* For the file open dialog */
#include "button.hpp"	/* The buttons */
#include "fileedit.hpp" /* For the file open dialog */
#include "text.hpp"		/* Text for prompt */
#include "event.hpp"	/* Critical Errors */
//** New includes
#include "editclus.hpp" /* Lists of editable strings */
#include "alert.hpp"	/* Confirm dialogs */

#define MAXLINE 256		/* Maximum input line */
#define MAXSEARCH 40	/* Maximum search string length */

// A class to hold phone numbers
// Derived from ListElement, which displays its name in a List
class Number : public ListElement {
private:
	char *k;	// The whole thing - "<cost code> <name> <number>"
	char *n;	// Points at the number part of k
	// initialise the variables given an input line
	// We keep this separate from the constructor, because we intend
	// to introduce more constructors as we expand the program.
	void build(const char *inpline);

public:
	// Construct from a file input line
	Number(const char *inpline);
	//** New constructor takes constituent parts
	Number(const char *key, const char *number, char code);
	// Tidy up allocated memory
	~Number();

	// Access methods.
	const char *key() const { return k + 2; }
	const char *number() const { return n; }
	char code() const { return *k; }
	};

// We are going to hold Numbers in a List, so we need a Collection
// to put them in.
declare(LCollection,Number);	// Something to hold a list of numbers
declare(LIterator,Number);		// An Iterator for the list
implement(LCollection,Number)

// We will allow several different types of search action
enum ScanMode {
	SMfind,		// Find the first key beginning with search string
	SMscan,		// Scan for the first entry containg the string anywhere
	SMnext,		// Repeat last search forwards
	SMprevious, // Repeat last search backwards
	SMcount		// The number of values for this enum (for array sizes)
	};

// We will allow several sort orders (well, two, anyway)
enum SortType {
	STname,		// Name order
	STnumber,	// Number order
	STcount		// The number of values for this enum (for array sizes)
	};

// Now lets derive a List class to hold the Numbers.  We could use
// the basic List class on its own, but deriving a class demonstrates
// how this should be done, and will make the program more extensible.
class PhoneList : public List {
private:
	LCollection(Number) directory;	// Collection of numbers from the file
	ScanMode scanmode;				// How we are searching this list
	char fname[OSPathLen+1];		// The file name
	int listchanged;				//** The list has been edited

public:
	// Constructor
	PhoneList(const char *filename = "PHONE.DIR");
	// Access method - return the currently selected number (or 0 if none)
	Number *selected() const { return directory[line()]; }
	// Read the List from a file. Use current file name if none supplied.
	int read(const char *filename = 0);
	// Find a string in the list
	int find(const char *string, ScanMode s = SMfind);
	// Force a complete re-display of the List
	void dirty() { changeCollection(); List::dirty(); }
	int openFile(const char *);		// Open file dialog
	int findNumber(const char *button);
	// Force the List to be as large as possible
	virtual coord maxheight() const;
	// Sort the List in a specified order
	void sort(SortType = STname);
	// Write the list back to file. Use current file name if none supplied.
	int write(const char *filename = 0);
	// Now the standard function that will be called by the sort Options
	int selectSort(const char *sortOption);
	// Standard function for the "Save" Option on the "File" menu
	int saveFile(const char *o);
	// Standard function for the "Save As" Option
	int saveAs(const char *o);
//** New methods
	// Mark the List as changed
	void changed() { changeCollection(); dirty(); listchanged = 1; }
	// Test if the List has changed
	int isdirty() const { return listchanged; }
	// Add a new Number to the List
	int add(Number *n);
	// Delete a Number from the List
	int del(Number *n);
// Ask the user whether to save the List if it has changed
	void checkSave();
// Now a standard type function called when the user presses the
// "Replace" (rather than "Add") button.
	int delNo(const char *o);
// Standard type function called when the user presses "Edit"
	int editNumber(const char *o);
// Standard type function for the "Delete" Button
	int delNumber(const char *o);
// Print the List
	int printFile(const char *o);
// The standard type function for the "Exit" Option on the "File" Menu
// must now be a member of PhoneList, as it calls a member function
	int exitProg(const char *o);
	};

declare(GCommand,PhoneList);
implement(GCommand,PhoneList)

// Now the constructor
Number::Number(const char *inpline)
: ListElement("")		// We will set the name to the data later
{
	while(isspace(*inpline))		// Skip leading spaces
		inpline++;

	char c = isalpha(*inpline) ? *inpline++ : ' ';	// Get the code

	while(isspace(*inpline))		// Skip any spaces between code and name
		inpline++;
	build(form("%c %s",c,inpline)); // The real "meat" of the constructor
}

//** The new constructor
Number::Number(const char *key, const char *number, char code)
: ListElement("")
{
	// Make up a dummy input line using form
	build(form("%c %s %s\n",code ? code : ' ',key,number));
}

// Destructor - delete the allocated memory
Number::~Number()
{
	newname("");		// we are about to delete name !
	delete k;
}

// The real construction work !
void Number::build(const char *inpline)
{
	int l = strlen(inpline);	// Length (including terminating newline)

	n = k = new char[l];		// Allocate the memory
	if(!k)
		return;					// failed !
	strncpy(k,inpline,l);		// Copy excluding terminating newline
	k[l - 1] = 0;				// Ensure null teminated
	for(n = k; *n && !isdigit(*n); n++)
		;						// Search along for the number
	newname(k);					// Tell the ListElement what to display
}

// Constructor
PhoneList::PhoneList(const char *filename)
// Pass the Collection, name, and border style to List
// The border will be a single line, the list will have a title,
// and the user will be allowed to Tab to the list.
: List(directory,strcpy(fname,filename),BorderFlag(BFsingle|BFtitle|BFtabstop)),
	listchanged(0)
{
	read();			// Read the data from the file
	sort(STname);	// And sort it
}

// Read the list from a file
int PhoneList::read(const char *filename)
{
	if(!filename)
		filename = fname;			// Use the default file name

	FILE *fd = fopen(filename,"r");

	if(fd == 0)
		return 0;					// Open failed - file not found ?
	directory.deleteall();			// Delete any existing contents

	static char inpline[MAXLINE+1]; // To hold the input lines.

	while(fgets(inpline,MAXLINE,fd))// Read lines from the file until EOF
		{
		if(inpline[0] == '\n' || inpline[0] == 0)
			continue;				// Skip blank lines
		directory.add(new Number(inpline)); // Add to Collection
		}
	fclose(fd);
	if(filename != fname)
		strcpy(fname,filename);		// Remember file name
	newname(fname);					// and set the title to it.
	dirty();						// Force a redraw
	listchanged = 0;				//** Mark list as unchanged
	return 1;
}

// We want to force the List to be as big as possible.
coord PhoneList::maxheight() const
{
	return 999;		// ask for as much space as possible.
}

// Write the List back to the file
int PhoneList::write(const char *filename)
{
	if(!filename)
		filename = fname;

	FILE *fd = fopen(filename,"w");		// create the file

	if(fd == 0)
		return 0;						// Error creating file

	LIterator(Number) next(directory);	// An Iterator for the Collection
	Number *c;

	while((c = next()) != 0)			// Traverse Collection
		fprintf(fd,"%c %s\n",c->code(),c->key());	// and write to file
	fclose(fd);							// Close file
	listchanged = 0;					//** Mark list as unchanged
	return 1;
}

// Now we declare the two sort functions (which are a standard type)
int alphasort(const Number *n1, const Number *n2)
{
#ifdef __ZTC__
	int retval = strcmpl(n1->key(),n2->key());		// ignore case
#endif
#ifdef __TURBOC__
	int retval = stricmp(n1->key(),n2->key());		// ignore case
#endif
	if(!retval)
		retval = strcmp(n1->key(),n2->key());		// don't ignore case
	return retval;
}

int numbersort(const Number *n1, const Number *n2)
{
	int retval = strcmp(n1->number(),n2->number());
	if(!retval)		// Numbers same - sort by name instead
		retval = alphasort(n1,n2);
	return retval;
}

// Typedef sort function pointer. I always do this as it makes
// the declaration of arrays of function pointers easier to read.
typedef int (*SORTFUNC)(const Number *,const Number *);
// An array of pointers to the sort functions.
SORTFUNC sortfunc[STcount] = {
	alphasort,		// Alphabetic order
	numbersort		// Numeric order
	};

// Now, the actual sort method
void PhoneList::sort(SortType i)
{
	Number *n = selected();			// Remember which one was selected

	deselect();						// and deselect it
	directory.sort(sortfunc[i]);	// Sort in order
	dirty();						// Force redisplay
	if(n)							// Reselect previously selected item
		select(directory.elementNo(n));
}
//** New methods
// Add a new Number to the List
int PhoneList::add(Number *n)
{
	int i = directory.add(n);	// Add to the Collection
					// Returns position in which new Number was added

	if(i < 0)		// Could not be added
		{
		delete n;	// So delete the new Number
		return 0;
		}
	changed();		// Mark List as changed
	select(i);		// Select newly added number
	makeVisible(i); // and ensure it is on screen
	return 1;
}

// Delete a Number from the List
int PhoneList::del(Number *n)
{
	if(!n || !directory.remove(n))		// Remove from Collection
		return 0;		// Wasn't in there to start with
	changed();			// Mark List as changed
	return 1;
}
//** End of new methods

// A utility function to find if a string starts with a pattern.
// Case is ignored in the search.
int strstart(const char *string, const char *pattern)
{
	for( ; *pattern; string++, pattern++)
		if(tolower(*string) != tolower(*pattern))
			return 0;		// Not found
	return 1;				// Found
}

// A utility function to find if a string contains a pattern anywhere in it.
int scan(const char *string, const char *pattern)
{
	int l = strlen(string) - strlen(pattern) + 1;

	for( ; l-- > 0; string++)
		if(tolower(*string) == tolower(*pattern)	// optimise a bit
			&& strstart(string,pattern))
				return 1;
	return 0;
}

// Search the PhoneList for a pattern
int PhoneList::find(const char *pattern, ScanMode s)
{
	Number *c = 0;			// Where to start the search
	int backwards = 0;		// Direction of search

	switch(s)
		{
	case SMfind:			// Search forwards
	case SMscan:			// Scan forwards
		scanmode = s;		// Remember search type for next time
		break;
	case SMprevious:
		backwards = 1;		// Search backwards
		/* drop through */
	case SMnext:			// Search forwards
		c = selected();		// From currently selected item
		}

	// We want to travrse the Collection, so declare an Iterator
	// Note that a simple forward search only requires
	// LIterator(Number) next(directory);
	LIterator(Number) next(directory,c,backwards);

	while((c = next()) != 0)			// Traverse the Collection
		{
		if(strstart(c->key(),pattern)	// Match or scan for pattern
										// depending on search mode
			|| (scanmode == SMscan && scan(c->key(),pattern)))
			{							// Found - move selection to element
			select(directory.elementNo(c));
			return 1;
			}
		}
	return 0;							// Not Found
}

// A standard function, called whenever the "Open"
// Option in the "File" Menu is selected.
int PhoneList::openFile(const char *)
{
	// We want to input a file name, so create a FileEdit
	FileEdit *f = new FileEdit(fname,"Number file");

	// A ModalDialog interacts with the user, then returns
	ModalDialog *dlg = new ModalDialog(
		// Lay out elements in a vertical Cluster
		new VCluster(ObjectCollection(2)
			+ f			 // The FileEdit
		// Now a horizontal Cluster of Buttons
			+ new HCluster(ObjectCollection(2)
		// A CloseButton which updates the FileEdit with the user's input
				+ new CloseEditButton(f)
		// and a CancelButton, in case the user changes their mind
				+ new CancelButton
				)
			),
		// The title of the Dialog
		"Open File");
	if(!dlg->enter())		// Interact with the user
		{
		checkSave();
		read();		// If "O.K." pressed, read the file
		}
	return 1;
}

char searchstring[MAXSEARCH+1];			// The pattern to find
const char *buttonNames[SMcount] =		// The different search buttons
	{ "F3=Find", "F4=Scan", "F5=Next", "F6=Previous" };

// Now, we need a function (of a standard type) to be called when the
// user presses a search Button. This function will be called automatically
// when one of the Buttons is pressed. Note that we have only one function
// here, and we look at the name of the Button which was pressed to
// find out which search strategy to use.  Aside from the advantage that
// we only have to write one function, it will not have to change if
// we add new types of search Button.
int PhoneList::findNumber(const char *button)
{
	for(int j = 0; j < SMcount; j++)
		if(!strcmp(button,buttonNames[j]))
			return find(searchstring,(ScanMode) j);
	return 0;
}

//** New Utility function
// Ask the user whether to save the List if it has changed
void PhoneList::checkSave()
{
	if(size() && isdirty())
		if(confirm("List has been amended\nsave to file ?"))
			write();
}

SortType sorttype;		// The current sort order

// Option names for the "View" Menu.  We are going to place an "X"
// next to the currently selected sort order.
char sortopts[STcount][30] = {
	"X Sort ~Alphabetically",
	"  Sort ~Numerically"
	};

// Now the standard function that will be called by the sort Options
int PhoneList::selectSort(const char *sortOption)
{
	// See which option was selected.
	for(int i = 0; i < STcount; i++)
		if(!strcmp(sortOption + 2,sortopts[i] + 2))
			{
			if(i == sorttype)
				return 1;					// Already in this order
			sortopts[sorttype][0] = ' ';	// Wipe X on previous order
			sortopts[i][0] = 'X';			// And set it on this order
			sorttype = (SortType) i;		// Remember sort type
			sort(sorttype);		 			// and do the sort
			break;
			}
	return 1;
}

// Standard function for the "Save" Option on the "File" menu
int PhoneList::saveFile(const char *)
{
	return write();
}

// Standard function for the "Save As" Option
int PhoneList::saveAs(const char *)
{
	// We want to input a file name.  The file does not have
	// to exist already.
	FileEdit *f = new FileEdit(fname,"Number file",1);
	// The Dialog
	ModalDialog *dlg = new ModalDialog(
		new VCluster(ObjectCollection(2)
			+ f		 // File name
			+ new HCluster(ObjectCollection(2)
				+ new CloseEditButton(f)
				+ new CancelButton
				)
			),
		"Save As"); // Dialog name
	if(!dlg->enter())
		write(); // Write back to specified file
	return 1;
}
//** New functions
// First, we want a TextEdit that doesn't allow digits, for inputting
// the name (remember, the first digit marks the start of the number)
// So we derive a new class from TextEdit
class KeyEdit : public TextEdit {
public:
	// Constructor
	KeyEdit(char *string,int length, const char *prompt = "");
	// Filter out digits
	virtual char filter(char inchar);
	};

// Constructor
KeyEdit::KeyEdit(char *string,int length, const char *prompt)
: TextEdit(string, length, prompt)
{
}

// Filter out digits
char KeyEdit::filter(char inchar)
{
	return isdigit(inchar) ? 0 : TextEdit::filter(inchar);
}

// Now a standard type function called when the user presses the
// "Replace" (rather than "Add") button.
int PhoneList::delNo(const char *)
{
	Number *n = selected();

	if(!n)
		return 0;			// Nothing currently selected
	del(n);		 			// Delete current selection
	return 1;
}

// Standard type function called when the user presses "Edit"
int PhoneList::editNumber(const char *)
{
	// Strings to contain parts of the Number
	static char newname[MAXSEARCH+1], newnumber[MAXSEARCH+1], newcode[2];
	// The currently selected (default) number
	Number *n = selected();

	if(n)				// Amend current selection
		{
		// Copy in the defaults
		strncpy(newname,n->key(),MAXSEARCH);
		strncpy(newnumber,n->number(),MAXSEARCH);
		newcode[0] = n->code();
		// The number() returns a pointer into key() which points
		// at the first digit.	We need to take this part off
		// the end of the name.
		int namelen = n->number() - n->key();

		if(namelen < strlen(newname))
			// Terminate the name at the first digit
			newname[namelen] = 0;
		}

	// Create a list of TextEdit's, which will be laid out
	// automatically.  The EditCluster also takes care of updating
	// the fields with the user's input when a Close or Apply Button
	// is pressed.
	EditCluster *e = new EditCluster(GCollection(EditControl)(3)
		+ new KeyEdit(newname,MAXSEARCH,"Name")
		+ new TextEdit(newnumber,MAXSEARCH,"Number")
		+ new UpperEdit(newcode,1,"Cost code")
		);
	// We will have two CloseButtons - "Add" and "Replace"
	// declare "Replace" here so we can disable it if there is
	// no current selection.
	CloseEditButton *replace = new CloseEditButton(e,new GCommand(PhoneList)(&PhoneList::delNo,this),notDefault,"Replace");
	// Now create the Add/Amend Dialog
	ModalDialog *dlg = new ModalDialog(
		new VCluster(ObjectCollection(2)
			+ e		 // The EditCluster declared above
			+ new HCluster(ObjectCollection(3)
				+ new CloseEditButton(e,"Add")
				+ replace
				+ new CancelButton,
				AlignFlag(AFexpand|AFtabstop))
			),
		// Dialog name depends on whether adding or amending
		n ? "Amend number" : "Add new number");

	if(!n)					// Adding
		replace->disable(); // so diable "Replace"
	if(!dlg->enter())		// Interact with the user
		add(new Number(newname,newnumber,newcode[0]));
	return 1;
}

// Standard type function for the "Delete" Button
int PhoneList::delNumber(const char *)
{
	Number *n = selected();

	if(!n)
		return 0;			// No selection to delete
	// Standard confirmation box - returns 1 for "Yes", 0 for "No"
	if(confirm(form("Delete entry\n%s",n->key())))
		del(n);
	return 1;
}

// Print the List
int PhoneList::printFile(const char *)
{
	write("LPT1");
	return 1;
}
//** End of new functions

// A standard type function for the "Exit" Option on the "File" Menu
int PhoneList::exitProg(const char *)
{
	checkSave();		// CHeck to see if List should be saved
	exit(0);
	return 1;
}

EventQueue eventQueue;	// We always have an EventQueue
Application example;	// We always have an Application

main()
{
	PhoneList *phonelist = new PhoneList();
	// Now an Edit to input the search string
	TextEdit *search = new TextEdit(searchstring,MAXSEARCH,"Search string");
	// and the first (default) "Find" button.  It must update searchstring
	// with the user's input, and it must call findNumber whenever it
	// is pressed.
	ApplyEditButton *find = new ApplyEditButton(search,
		buttonNames[SMfind],KF3,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist));

	find->makedefault();		// Make this the default Button
								// (selected when the user presses Return)
	// and the Panel
	Panel *main = new Panel(
	// Panels always have a MenuBar
		new MenuBar(KF10,			// This is the SAA standard hot-key
			OCollection(MenuButton)(3)
	// The first Menu in the MenuBar
			+ new MenuButton("~File",KAltF,LCollection(Option)(5)
				+ new Option("~Open",new GCommand(PhoneList)(&PhoneList::openFile,phonelist))
				+ new Option("~Save",new GCommand(PhoneList)(&PhoneList::saveFile,phonelist))	//** New Option
				+ new Option("Save ~As",new GCommand(PhoneList)(&PhoneList::saveAs,phonelist))	//** New Option
				+ new Option("~Print",new GCommand(PhoneList)(&PhoneList::printFile,phonelist)) //** New Option
				+ new Option("~Exit",new GCommand(PhoneList)(&PhoneList::exitProg,phonelist),KAltF4)
				)
			+ new MenuButton("~View",KAltV,LCollection(Option)(2)
				+ new Option(sortopts[STname],new GCommand(PhoneList)(&PhoneList::selectSort,phonelist))
				+ new Option(sortopts[STnumber],new GCommand(PhoneList)(&PhoneList::selectSort,phonelist))
				)
	// Allow context-sensitive help
			+ DefaultHelpMenu
			),
	// Now the client area, which is arranged vertically
		new VCluster(ObjectCollection(3)
	// the search string prompt, and field
			+ new HCluster(ObjectCollection(2)
				+ new FixedText("Search for ")
				+ search
				)
	// The search Buttons
			+ new HCluster(ObjectCollection(6)
				+ find	 // we defined this one earlier
	// Here are the others
				+ new ApplyEditButton(search,
					buttonNames[SMscan],KF4,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist))
				+ new ApplyEditButton(search,
					buttonNames[SMnext],KF5,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist))
				+ new ApplyEditButton(search,
					buttonNames[SMprevious],KF6,new GCommand(PhoneList)(&PhoneList::findNumber,phonelist))
	//** New Buttons
	// We don't need to update the search string for these,
	// so we use simple ActiveButtons
				+ new ActiveButton(new GCommand(PhoneList)(&PhoneList::editNumber,phonelist),"F7=Edit",KF7)
				+ new ActiveButton(new GCommand(PhoneList)(&PhoneList::delNumber,phonelist),"F8=Delete",KF8),
				"Search list",AFHexpand,BorderFlag(BFsingle|BFtabstop))
	// The PhoneList itself
			+ phonelist
			),
	// Panel name
		"Phone numbers");
	main->show();		// Place the Panel on the screen
	//** Trap DOS critical errors - it really is as simple as this !
	//** If a critical error occurrs, an alert dialog will appear
	//** detailing the error, and the error will be returned to
	//** DOS (Fail)
	CriticalErrorHandler catchErrors;	// Trap DOS Critical Errors
	// Now open the help file. This is all you have to do to add context
	// sensitive help to an application.  Of course, you have to
	// write the actual help text - see "PHONE.SRC" for details.
	Application::openHelpFile("phone");
	example.enter();	// Enter the main message loop
	return 0;
}


