#ifndef LIST_HPP
#define LIST_HPP
#ifndef VOBJECT_HPP
#include "vobject.hpp"
#endif
#ifndef COLLECT_HPP
#include "collect.hpp"
#endif

class Window;

/*
A typical element of a List. This class is provided as a building
block, to be subclassed for anything that needs to be displayed in a
List. When it is painted, it merely shows its name. The default
width is the length of its name. 
*/
class ListElement {
	const char *namestring;
public:
	ListElement(const char *name);
	// The pointer itself is held - the name is not saved in allocated storage.
	virtual ~ListElement();
	const char *name() const { return namestring; }
	void newname(const char *n) { namestring = n; }
	// Change the name.
	virtual void paint(const Rectangle &at, Window *w, const Rectangle &clip, 
		color col);
	virtual int minwidth() const;
	virtual int maxwidth() const;
	};

/*
This is the base class on which the generic collections of ListElements
are based. It allows sorting, so that Lists can be displayed in any
sort order required.
See LCollection(type)
*/
class ListCollection : public Collection {
	int (*compfn)(const ListElement *,const ListElement *); // comparison function
public:
	ListCollection(int hintSize = 0);
	ListCollection(const ListCollection& c);
	ListCollection& operator= (const ListCollection& c) 
		{ Collection::operator=(c); compfn = c.compfn; return *this; } 
	ListCollection& operator+ (const ListCollection &c)
		{ Collection::operator+(c); return *this; }
	ListCollection& operator+ (const ListCollection *c)
		{ Collection::operator+(c); return *this; }
	ListCollection& operator+ (ListElement *c) { add(c); return *this; }
	int add(ListElement *item);
	void insert(ListElement *item) { Collection::insert(item); }
	void append(ListElement *item) { Collection::append(item); }
	int remove(const ListElement *item) { return Collection::remove(item); }
	void deleteall();
	ListElement *first() const { return (ListElement *)Collection::first(); }
	ListElement *last() const { return (ListElement *)Collection::last(); }
	ListElement *operator[](int i) const 
		{ return (ListElement *)Collection::operator[](i); }
	int elementNo(const ListElement *e) const { return Collection::elementNo(e); }
	ListElement *get() { return (ListElement *)Collection::get(); }
	// Remove and return the first item from the Collection.
	ListElement *getlast() { return (ListElement *)Collection::getlast(); }
	// Remove and return the last item from the Collection.
	// This is more efficient than get, because of the way 
	// Collections are implemented.
	void sort(int (*compare)(const ListElement *,const ListElement *))
	// Sort the list, using the compare function provided.
		{ compfn = compare; resort(); }
	virtual int compare(const ListElement *v1, const ListElement *v2) const;
	// Compare two elements using the current sort function.
	void resort();
};

/*
Iterate through a ListCollection. See LIterator for generic subclass.
*/
class ListIterator : public Iterator {
public:
	ListIterator(const ListCollection& s,int bk = 0);
	ListIterator(const ListCollection& s,const ListElement *e,int bk = 0);
	int set(const ListElement *s) { return Iterator::set(s); }
	ListElement *next() { return (ListElement *)Iterator::next(); }
	ListElement *previous() { return (ListElement *)Iterator::previous(); }
	ListElement *operator()()
		{ return (ListElement *)Iterator::operator()(); }
};

#define LCollection(type) name2(LCollection,type)

/*
Generic collection of ListElements.
It allows sorting, so that Lists can be displayed in any sort order
required.To make a generic collection of (say) MyListElements (derived
from ListElement), in your header (.hpp) file do ... 

declare(LCollection, MyListElement);

and in your .cpp file do ...

implement(LCollection, MyListElement)

(Note the lack of a semi-colon).
*/
#define LCollectiondeclare(type)\
class LCollection(type): public ListCollection {\
	int (*compfn)(const type *,const type *);\
public:\
	LCollection(type)(int hintSize = 0);\
	LCollection(type)(const LCollection(type)& c);\
	LCollection(type)& operator= (const LCollection(type)& c) \
		{ ListCollection::operator=(c); compfn = c.compfn; return *this; } \
	LCollection(type)& operator+ (const LCollection(type) &c) \
		{ ListCollection::operator+(c); return *this; } \
	LCollection(type)& operator+ (const LCollection(type) *c) \
		{ ListCollection::operator+(c); return *this; } \
	LCollection(type)& operator+ (type *c) { add(c); return *this; }\
	int add(type *item) { return ListCollection::add(item); }\
	void insert(type *item) { ListCollection::insert(item); }\
	void append(type *item) { ListCollection::append(item); }\
	int remove(const type *item) { return ListCollection::remove(item); }\
	type *first() const { return (type *)ListCollection::first(); }\
	type *last() const { return (type *)ListCollection::last(); }\
	type *operator[](int i) const \
		{ return (type *)ListCollection::operator[](i); }\
	int elementNo(const type *e) const { return ListCollection::elementNo(e); }\
	type *get() { return (type *)ListCollection::get(); }\
	type *getlast() { return (type *)ListCollection::getlast(); }\
	void sort(int (*compare)(const type *,const type *))\
		{ compfn = compare; resort(); }\
	virtual int compare(const ListElement *v1, const ListElement *v2) const;\
}

#define LCollectionimplement(type)\
LCollection(type)::LCollection(type)(int hintSize)\
: ListCollection(hintSize), compfn(0) {}\
LCollection(type)::LCollection(type)(const LCollection(type)& c)\
: ListCollection(c), compfn(0) {} \
int LCollection(type)::compare(const ListElement *v1, const ListElement *v2) const\
{\
	return compfn ? compfn((type *)v1,(type *)v2) : 0;\
}

#define LIterator(type) name2(LIterator,type)

/*
Generic Iterator over LCollection.
To make a generic Iterator for a Collection of (say) MyListElements,
in your header (.hpp) file do ... 

declare(LIterator, MyListElement);

(Note - no implementation is necessary, as all methods are inline).
*/
#define LIteratordeclare(type)\
class LIterator(type) : public ListIterator {\
public:\
	LIterator(type)(const LCollection(type)& s,int bk = 0) : ListIterator(s,bk) {}\
	LIterator(type)(const LCollection(type)& s,const type *e,int bk = 0) \
		: ListIterator(s,e,bk) {}\
	int set(const type *s) { return ListIterator::set(s); }\
	type *next() { return (type *)ListIterator::next(); }\
	type *previous() { return (type *)ListIterator::previous(); }\
	type *operator()()\
		{ return (type *)ListIterator::operator()(); }\
}

/*
The basic displayable List. The members of the list are ListElements.
It is used for pop-down lists, lists of Menu Options etc. 
The ListCollection parameter to the constructors should be a reference 
to some LCollection (usually held in the derived class).
*/
class List : public Control {
public:
	List(ListCollection &c, const char *n = "List",
		const BorderFlag b = BFsingle);
	// The size is calculated from the contents.
	List(ListCollection &c,coord width,coord height,
		const char *n = "List",const BorderFlag b = BFsingle);
	// The size is specified in the constructor.

	int line() const { return selected; }
	// The item number of the currently selected item, or -1.
	int size() const { return collection.size(); }
	// The number of items in the list.
	virtual int select(const int s);
	// Select an item in the list. Returns 1 if s was a valid item number.
	virtual void deselect();		// Deselect the current item (if any).
	virtual int pick();
	// This virtual method is called whenever an item is selected with <cr>
	// or click. It may be overridden in subclasses - the default behaviour
	// is to ignore the event. 
	virtual int escape();			// Called after Esc or rightclick.
	// This virtual method is called when the user presses Esc or clicks
	// the right mouse button over the list. Again, the default behaviour
	// is to ignore the event. 
	int topitem() const { return topItem; }
	// Returns the item number of the top visible line.
	void deleteall() { collection.deleteall(); changeCollection(); }
	// Deletes everything in the list.
	void changeCollection();		// Inform list of a change to collection.
	int remove(int itemno);			// Remove item from collection.
	int add(ListElement *o);
	// Add item to collection.
	// For type safety, this should be overridden in derived Lists.

	virtual void setheight(const coord w);
	virtual coord minwidth() const;
	virtual coord minheight() const;
	virtual coord maxwidth() const;
	virtual coord maxheight() const;

	virtual void paint(Rectangle r);
	// redraw part of the Object

	virtual int setfocus(int tabstop = 1);
	virtual void resetfocus();

	virtual int leftdown(MouseEvent& e);
	virtual int leftup(MouseEvent& e);
	virtual int leftdoubleclick(MouseEvent& e);

	virtual int rightclick(MouseEvent& e);
	virtual int mousemove(MouseEvent& e);
	virtual int mouseexit(MouseEvent& e);

	virtual int kbdevent(KeyboardEvent& e);

	virtual char *inspectinfo(char *buffer = 0,const char *s = 0) const/*!*/;
protected:
	BorderFlag bflags;
	ListCollection &collection;
	int selected;			// Item no of selected item (or -1).
	int topItem;			// Item no of top visible item in collection.

	void resize(int changeheight = 0);
	// Make the list the right size for the collection. The width of the
	// List will be the size of the widest ListElement in it (plus the
	// border), and the size of the list will be reduced so there are no
	// empty slots visible. If changeheight is non-zero, the height will be
	// increased if the collection can't all be seen. 
	void scrollup(int nlines = 1);	// Show earlier items in collection.
	void scrolldown(int nlines = 1);// Show later items in collection.
	void dirtyItem();				// Mark selected item as dirty.
	void makeVisible(int item);		// Ensure item is in visible part.
	};

#endif
