// Menus - bar and pop-down.
#ifndef MENU_HPP
#define MENU_HPP
#ifndef LIST_HPP
#include "list.hpp"
#endif
#ifndef CLUSTER_HPP
#include "cluster.hpp"
#endif
#ifndef POPWIN_HPP
#include "popwin.hpp"
#endif

class Command;

/*
A Menu Option.
The name of the Option may contain a tilde (~) character to
indicate that the following character is the quick selection
key for the Option. The tilde is not shown on the screen, and
the quick key is highlighted. If hotkey is set to Kauto, then the 
jotkey for that Option will automatically be set to Alt+the quick 
key.
*/
class Option : public ListElement {
	void build(const char *text, Keys hotkey);
protected:
	int disabled;
public:
	Option(const char *text,Command *command,Keys hotkey = Knone);
	// The Command is executed when the Option is selected.
	Option(const char *text,int (*func)(const char *),Keys hotkey = Knone);
	// The func is called when the Option is selected.
	~Option();

	void on(int switchon = 1)
	// Enable and disable the Option. The option appears greyed when
	// it is disabled. 
		{ disabled = !switchon; }
	void off() { on(0); }

	virtual int doit();

	virtual coord minwidth() const;

	virtual void paint(const Rectangle &at, Window *w, const Rectangle &clip, 
		color col);

	int quickkey() const { return quick; }
	Keys hotkey() const { return hot; }

private:
	Command *cmd;
	Keys hot;
	int quick;
	};

declare(LCollection,Option);
declare(LIterator,Option);

/*
A special two-state Option which indicates its state with the 
presence or absence of a tick to the left of the Option name.
*/
class TickOption : public Option {
	int &tick;
public:
	TickOption(const char *text, int &result, Keys hotkey = Knone,
		Command *command = 0);
	// Result represents the state. The initial tick state is set if
	// result is non-zero. As the Option is toggled on/off, the result
	// is set to 1/0.
	virtual int doit();
	virtual coord minwidth() const;
	virtual void paint(const Rectangle &at, Window *w, const Rectangle &clip, 
		color col);
	};

extern ColorSet defaultColors;		// Used when no specific colors specified
class MenuButton;

/*
A pop-down menu. Menus are normally attached to MenuButtons, and are
invoked automatically when the button is pressed. 
*/
class Menu : public List {
public:
// Vararg constructors are obsolete
	Menu(const char *name,const ColorSet &c,Option *o,...)/*!*/;
	Menu(const char *name,Option *o,...)/*!*/;

//~ Recommended constructors
	Menu(const char *name,const ColorSet &c = defaultColors);
	Menu(const char *name,const LCollection(Option) &opts,
		const ColorSet &c = defaultColors);

//
	~Menu();

	void setMenuColor(const ColorSet &c) { if(popwin) popwin->setcolors(c); }

	Menu& operator+ (const LCollection(Option) &c);
	Menu& operator+ (const LCollection(Option) *c)
		{ operator+(*c); return *this; }
	Menu& operator+ (Option *o) { addOption(o); return *this; }
	void addOption(Option *o);

	void addvalist(Option *o,va_list args)/*!*/;

	virtual int pick();

	void pop(VObject *v);

	virtual void resetfocus();

	virtual int leftup(MouseEvent& e);
	virtual int rightclick(MouseEvent &e);

	virtual int kbdevent(KeyboardEvent& e);
	virtual int dohotkey(KeyboardEvent& e);
	virtual int help();

private:
	LCollection(Option) options;
	class PopWindow *popwin;
	MenuButton *kbdhandler;
friend class MenuButton;
	};

class OCollectionMenuButton/*!*/;

/*
A menu bar, consisting of MenuButtons, each with its own Menu. A
MenuBar is created automatically when you create a Panel. 
*/
class MenuBar : public HCluster {
public:
	MenuBar(Keys hotkey,MenuButton *b,...)/*!*/;
	MenuBar(Keys hotkey, const OCollectionMenuButton &c);
	// SAA convention says the hotkey should be KF10.

	void menustarted() { setFlag(SFselected); }// Called when a Menu appears.
	void menudone() { clearFlag(SFselected); } // Called when a Menu disappears.
	
	void setMenuColors(const ColorSet &c);

	virtual coord maxwidth() const;

	virtual void paint(Rectangle r);

	virtual int kbdevent(KeyboardEvent& e);
	virtual int dohotkey(KeyboardEvent& e);

private:
	Keys hot;
	};

/*
A Button which pops down a Menu when it is pushed. 
The name may contain one tilde (~) character just before a letter or
digit. This will be used as the quick key - pressing this key while
the MenuBar containing this MenuButton is in focus will select the
button, and pop down the Menu. The default value of hotkey (Kauto) will
cause this same key to be used with the Alt key as the hotkey for
this MenuButton. Pressing Alt+this key will immediately pop down the
Menu, from anywhere in the Application. 

DefaultHelpMenu is a #define macro which evaluates to a MenuButton 
which will show the standard SAA Help menu.
*/
class MenuButton : public Control {
public:
	MenuButton(const char *n,Menu *m,Keys hotkey = Kauto);
	MenuButton(const char *n,Keys hotkey,Option *o,...)/*!*/;
	MenuButton(const char *n,Keys hotkey,const LCollection(Option)& c);
	~MenuButton();

	void setMenuColor(const ColorSet &c) { if(menu) menu->setMenuColor(c); }
	void menudone()		// Called when a Menu disappears.
		{ if(kbdhandler) kbdhandler->menudone(); }
	void unfocus()		// Called from Menu::resetfocus()
		{ clearFlag(SFfocus); dirty(); }
	void popmenu();

	virtual void contentsDeleted(VObject* v);

	int setpressed();

	virtual int setfocus(int tabstop = 1);

	virtual coord minwidth() const;

	virtual void paint(Rectangle r);

	virtual int leftdown(MouseEvent& e);
	virtual int kbdevent(KeyboardEvent& e);
	virtual int dohotkey(KeyboardEvent& e);
	int quickkey() const { return quick; }
	Keys hotkey() const { return hot; }

private:
	Keys hot;
	int quick;
	Menu *menu;
	MenuBar *kbdhandler;
friend class MenuBar;
	};
declare(OCollection,MenuButton);
declare(OIterator,MenuButton);

/*
Call this functuion from anywhere in your application to show the 
page of help with the specified name, or the help index if no such 
help can be found.
*/
extern int showHelpMenu(const char *helpkey);
#define DefaultHelpMenu new MenuButton("Help",Knone,\
	new Option("Help for help",showHelpMenu),\
	new Option("Keys help",showHelpMenu),\
	new Option("Help index",showHelpMenu),\
	EndList)

#endif

