// Protection for classes.
#ifndef PROTECT_HPP
#define PROTECT_HPP
#ifndef HARDWARE_HPP
#include "hardware.hpp"
#endif
#include <stdio.h>

#ifndef NOPROTECT
#include <assert.h>

#define VALIDATE(p) (p->valid() ? (void)0 : \
					crash("Invalid object", __FILE__, __LINE__))

/*
This class forms the basis for the built-in object validity checking
system and debugging tool. All the overhead of using protection can
be turned off by recompiling the library with NOPROTECT defined. In
the checking implementation, any object of a class derived from
Protected can be validated by calling its valid method. An assertion
that an object is valid can be placed in the code using the VALIDATE
macro. 

Objects of this class register themselves into a debugging list (one
for new'ed items, one for static & automatic) which can be dumped.
Objects "know" whether they were allocated with new, and also
whether they have been deleted, or their memory reused. 
*/
class Protected {
private:
  Protected *me;
  Protected *next;          // Linked list next pointer
  void *caller;             // Address of calling function
  int newflag;              // != 0 if created with new
  static void *lastnew;     // Last return value from operator new
  static size_t length;     // Size of last new'ed item
  static Protected *pstack; // Static/automatic list
  static Protected *pheap;  // New (heap) list
  virtual void v_dump(FILE *f) const;
                            // Print details of this object
  void setNewFlag();        // called by constructors to set newflag & caller
public:
  Protected();
  Protected(const Protected &); // Prevent copying of members.
  ~Protected();
  Protected &operator=(const Protected &)
    { return *this; }           // Prevent copying of members.
  int valid() const { return (this == me); }
  // Test whether the object is still valid (i.e. hasn't been deleted)
  int isnew() const { return newflag; }
  // Test whether the object was allocated with new.
  void dump(FILE *f) const 
  // If the object is valid, dump its details to the file.
    { VALIDATE(this); v_dump(f ? f : dumpfile); }
  void *operator new(size_t);
  // Operator new is overridden so that a Protected object can
  // determine whether it was created with new.
  void *calledFrom() const;
  static void dumpstack(FILE *f = 0);
  // Dump all valid objects not allocated with new.
  static void dumpheap(FILE *f = 0);
  // Dump all valid objects allocated with new.
  static void dumpall(FILE *f = 0) { dumpstack(f); dumpheap(f); }
  // Dump all valid objects.
  static void dumpclose(FILE *f = 0);
  static const char on;
/*
This static constant is used to ensure modules compiled with a
mixture of protection on and off cannot be linked together by
mistake. This is taken care of automatically whenever protect.hpp is
included - each module that includes it will contain a static
constructor with a reference to either Protect::on or Protect::off,
depending whether protection was on or off. This facility can be
turned off by defining LIBRARY_MODULE. This is most useful when you
have a single source file which is compiled to give multiple object
files (like the library source files), so you cannot have static
constructors in all the object files. 
*/
  static FILE *dumpfile;
  // Set to an open file to enable automatic dumping on a crash.
  };
#ifndef LIBRARY_MODULE
static const char protectionCheck = Protected::on /*!*/;
#endif

// A Protected item with a name (for dumping).
class NamedProtected : public Protected {
  virtual void v_dump(FILE *f) const;
  // New version of dump prints name
protected:
  const char *namestring;
public:
  NamedProtected(const char *name);
  const char *name() const { return namestring; }
  };

// A named function tag with file and line number.
class Function : public NamedProtected {
  virtual void v_dump(FILE *f) const;
  // Prints name, file and line number
  const char *filename;
  unsigned lineno;
public:
  Function(const char *name, unsigned &frame, const char *file, unsigned line);
  const char *file() const { return filename; }
  unsigned line() const { return lineno; }
  };
#define FUNCTION(name) unsigned uit__temp__; Function(name,uit__temp__,__FILE__,__LINE__);
#else
//!
// No checking, no overhead version
class Protected {
public:
  int valid() const { return 1; }
  static void dumpstack(FILE *f = 0) {}
  static void dumpheap(FILE *f = 0) {}
  static void dumpall(FILE *f = 0) {}
  static void dumpclose(FILE *f = 0) {}
  static const char off;  // prevent linking mixed versions
  };

class NamedProtected : public Protected {
protected:
  const char *namestring;
public:
  NamedProtected(const char *name);
  const char *name() const { return namestring; }
  };

#define FUNCTION(ignore) ((void)0)
#define VALIDATE(p) ((void)0)
#ifndef LIBRARY_MODULE
static const char protectionCheck = Protected::off;
#endif
#endif

#endif
